"""Feature Schemas used by various Afero resources."""

from dataclasses import dataclass, field
from enum import Enum

from ...util import percentage_to_ordered_list_item


@dataclass
class ColorModeFeature:
    """Represent the current mode (ie white, color) Feature object"""

    mode: str

    @property
    def api_value(self):
        return self.mode


@dataclass
class ColorFeature:
    """Represent `RGB` Feature object"""

    red: int
    green: int
    blue: int

    @property
    def api_value(self):
        return {
            "value": {
                "color-rgb": {
                    "r": self.red,
                    "g": self.green,
                    "b": self.blue,
                }
            }
        }


@dataclass
class ColorTemperatureFeature:
    """Represent Current temperature Feature"""

    temperature: int
    supported: list[int]
    prefix: str | None = None

    @property
    def api_value(self):
        return f"{self.temperature}{self.prefix}"


class CurrentPositionEnum(Enum):
    """Enum with available current position modes."""

    LOCKED = "locked"
    LOCKING = "locking"
    UNKNOWN = "unknown"
    UNLOCKED = "unlocked"
    UNLOCKING = "unlocking"

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


@dataclass
class CurrentPositionFeature:
    """Represents the current position of the lock"""

    position: CurrentPositionEnum

    @property
    def api_value(self):
        return self.position.value


@dataclass
class DimmingFeature:
    """Represent Current temperature Feature"""

    brightness: int
    supported: list[int]

    @property
    def api_value(self):
        return self.brightness


@dataclass
class DirectionFeature:
    """Represent Current Fan direction Feature"""

    forward: bool

    @property
    def api_value(self):
        return "forward" if self.forward else "reverse"


@dataclass
class EffectFeature:
    """Represent the current effect"""

    effect: str
    effects: dict[str, set[str]]

    @property
    def api_value(self):
        states = []
        seq_key = None
        for effect_group, effects in self.effects.items():
            if self.effect not in effects:
                continue
            else:
                seq_key = effect_group
                break
        preset_val = self.effect if self.effect in self.effects["preset"] else seq_key
        states.append(
            {
                "functionClass": "color-sequence",
                "functionInstance": "preset",
                "value": preset_val,
            }
        )
        if seq_key != "preset":
            states.append(
                {
                    "functionClass": "color-sequence",
                    "functionInstance": seq_key,
                    "value": self.effect,
                }
            )
        return states

    def is_preset(self, effect):
        try:
            return effect in self.effects["preset"]
        except KeyError:
            return False


@dataclass
class HVACModeFeature:
    """Represent HVAC Mode Feature"""

    mode: str | None
    previous_mode: str | None
    supported_modes: set[str]
    modes: set[str]

    @property
    def api_value(self):
        return self.mode


@dataclass
class ModeFeature:
    """Represent Current Fan mode Feature"""

    mode: str | None
    modes: set[str]

    @property
    def api_value(self):
        return self.mode


@dataclass
class NumbersFeature:
    """Represents a numeric value"""

    value: float
    min: float
    max: float
    step: float
    name: str | None
    unit: str | None

    @property
    def api_value(self):
        return self.value


@dataclass
class OnFeature:
    """Represent `On` Feature object as used by various Afero resources."""

    on: bool
    func_class: str | None = field(default="power")
    func_instance: str | None = field(default=None)

    @property
    def api_value(self):
        state = {
            "value": "on" if self.on else "off",
            "functionClass": self.func_class,
        }
        if self.func_instance:
            state["functionInstance"] = self.func_instance
        return state


@dataclass
class OpenFeature:
    """Represent `Open` Feature object"""

    open: bool
    func_class: str | None = field(default="toggle")
    func_instance: str | None = field(default=None)

    @property
    def api_value(self):
        state = {
            "value": "on" if self.open else "off",
            "functionClass": self.func_class,
        }
        if self.func_instance:
            state["functionInstance"] = self.func_instance
        return state


@dataclass
class PresetFeature:
    """Represent the current preset"""

    enabled: bool
    func_instance: str
    func_class: str

    @property
    def api_value(self):
        return {
            "functionClass": self.func_class,
            "functionInstance": self.func_instance,
            "value": "enabled" if self.enabled else "disabled",
        }


@dataclass
class SelectFeature:
    """Represent available options and currently selected"""

    selected: str
    selects: set[str]
    name: str

    @property
    def api_value(self):
        return self.selected


@dataclass
class SpeedFeature:
    """Represent Current Fan speed Feature"""

    speed: int
    speeds: list[str]

    @property
    def api_value(self):
        return percentage_to_ordered_list_item(self.speeds, self.speed)


@dataclass
class TargetTemperatureFeature:
    """Represents the target temperature for auto"""

    value: float
    min: float
    max: float
    step: float
    instance: str

    @property
    def api_value(self):
        return {
            "functionClass": "temperature",
            "functionInstance": self.instance,
            "value": self.value,
        }


AferoFeatures: list = [
    ColorModeFeature,
    ColorFeature,
    ColorTemperatureFeature,
    CurrentPositionEnum,
    CurrentPositionFeature,
    DimmingFeature,
    DirectionFeature,
    EffectFeature,
    HVACModeFeature,
    ModeFeature,
    NumbersFeature,
    OnFeature,
    OpenFeature,
    PresetFeature,
    SelectFeature,
    SpeedFeature,
    TargetTemperatureFeature,
]



"""Test FanController"""

import asyncio
import logging

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers import event
from aioafero.v1.controllers.fan import FanController, features

from .. import utils

zandra_fan = utils.create_devices_from_data("fan-ZandraFan.json")[0]


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    controller = FanController(mocked_bridge)
    yield controller


@pytest.mark.asyncio
async def test_initialize_zandra(mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "066c0e38-c49b-4f60-b805-486dc07cab74"
    assert dev.on == features.OnFeature(on=True)
    assert dev.speed == features.SpeedFeature(
        speed=50,
        speeds=[
            "fan-speed-6-016",
            "fan-speed-6-033",
            "fan-speed-6-050",
            "fan-speed-6-066",
            "fan-speed-6-083",
            "fan-speed-6-100",
        ],
    )
    assert dev.direction == features.DirectionFeature(forward=False)
    assert dev.device_information.model == "Zandra"


@pytest.mark.asyncio
async def test_turn_on(mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = False
    await mocked_controller.turn_on(zandra_fan.id)
    req = utils.get_json_call(mocked_controller)
    assert req["metadeviceId"] == zandra_fan.id
    expected_states = [
        {
            "functionClass": "power",
            "functionInstance": "fan-power",
            "lastUpdateTime": 12345,
            "value": "on",
        }
    ]
    utils.ensure_states_sent(mocked_controller, expected_states)


@pytest.mark.asyncio
async def test_turn_off(mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    assert len(mocked_controller.items) == 1
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = True
    await mocked_controller.turn_off(zandra_fan.id)
    req = utils.get_json_call(mocked_controller)
    assert req["metadeviceId"] == zandra_fan.id
    expected_states = [
        {
            "functionClass": "power",
            "functionInstance": "fan-power",
            "lastUpdateTime": 12345,
            "value": "off",
        }
    ]
    utils.ensure_states_sent(mocked_controller, expected_states)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "speed, expected_speed",
    [
        # Speed of 0 should turn off
        (0, None),
        # Find the next highest value
        (1, "fan-speed-6-016"),
        # Exact value
        (16, "fan-speed-6-016"),
    ],
)
async def test_set_speed(speed, expected_speed, mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    if expected_speed:
        dev.on.on = False
    else:
        dev.on.on = True
    dev.speed.speed = "fan-speed-6-100"
    await mocked_controller.set_speed(zandra_fan.id, speed)
    req = utils.get_json_call(mocked_controller)
    assert req["metadeviceId"] == zandra_fan.id
    expected_states = [
        {
            "functionClass": "power",
            "functionInstance": "fan-power",
            "lastUpdateTime": 12345,
            "value": "on",
        },
    ]
    if expected_speed is None:
        expected_states[0]["value"] = "off"
    else:
        expected_states.append(
            {
                "functionClass": "fan-speed",
                "functionInstance": "fan-speed",
                "lastUpdateTime": 12345,
                "value": expected_speed,
            }
        )
    utils.ensure_states_sent(mocked_controller, expected_states)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "on",
    [True, False],
)
@pytest.mark.parametrize(
    "forward, value",
    [
        (True, "forward"),
        (False, "reverse"),
    ],
)
async def test_set_direction(on, forward, value, mocked_controller, caplog):
    caplog.set_level(logging.DEBUG)
    await mocked_controller.initialize_elem(zandra_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = on
    dev.direction.forward = not forward
    await mocked_controller.set_direction(zandra_fan.id, forward)
    req = utils.get_json_call(mocked_controller)
    assert req["metadeviceId"] == zandra_fan.id
    expected_states = [
        {
            "functionClass": "fan-reverse",
            "functionInstance": "fan-reverse",
            "lastUpdateTime": 12345,
            "value": value,
        },
    ]
    if not on:
        assert "Fan is not running so direction will not be set" in caplog.text
    utils.ensure_states_sent(mocked_controller, expected_states)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "on",
    [
        (True,),
        (False,),
    ],
)
@pytest.mark.parametrize(
    "preset, value",
    [
        (True, "enabled"),
        (False, "disabled"),
    ],
)
async def test_set_preset(on, preset, value, mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    dev.on.on = True
    dev.preset.enabled = not preset
    await mocked_controller.set_preset(zandra_fan.id, preset)
    req = utils.get_json_call(mocked_controller)
    assert req["metadeviceId"] == zandra_fan.id
    expected_states = [
        {
            "functionClass": "toggle",
            "functionInstance": "comfort-breeze",
            "lastUpdateTime": 12345,
            "value": value,
        },
    ]
    utils.ensure_states_sent(mocked_controller, expected_states)


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("fan-ZandraFan.json")[0]
    new_states = [
        AferoState(
            **{
                "functionClass": "toggle",
                "value": "disabled",
                "lastUpdateTime": 0,
                "functionInstance": "comfort-breeze",
            }
        ),
        AferoState(
            **{
                "functionClass": "fan-speed",
                "value": "fan-speed-6-016",
                "lastUpdateTime": 0,
                "functionInstance": "fan-speed",
            }
        ),
        AferoState(
            **{
                "functionClass": "fan-reverse",
                "value": "forward",
                "lastUpdateTime": 0,
                "functionInstance": "fan-reverse",
            }
        ),
        AferoState(
            **{
                "functionClass": "power",
                "value": "off",
                "lastUpdateTime": 0,
                "functionInstance": "fan-power",
            }
        ),
        AferoState(
            **{
                "functionClass": "toggle",
                "value": "disabled",
                "lastUpdateTime": 0,
                "functionInstance": "comfort-breeze",
            }
        ),
        AferoState(
            **{
                "functionClass": "available",
                "value": False,
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.preset.enabled is False
    assert dev.speed.speed == 16
    assert dev.direction.forward is True
    assert dev.on.on is False
    assert dev.available is False
    assert updates == {"speed", "direction", "on", "preset", "available"}


@pytest.mark.asyncio
async def test_update_elem_no_updates(mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    assert len(mocked_controller.items) == 1
    updates = await mocked_controller.update_elem(zandra_fan)
    assert updates == set()


# @TODO - Create tests for BaseResourcesController
@pytest.mark.asyncio
async def test_update(mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.on.on is True
    manual_update = [
        {
            "functionClass": "power",
            "value": "off",
            "functionInstance": "fan-power",
        }
    ]
    await mocked_controller.update(zandra_fan.id, states=manual_update)
    assert dev.on.on is False


@pytest.mark.asyncio
async def test_set_state_empty(mocked_controller):
    await mocked_controller.initialize_elem(zandra_fan)
    await mocked_controller.set_state(zandra_fan.id)


@pytest.mark.asyncio
async def test_fan_emitting(bridge):
    dev_update = utils.create_devices_from_data("fan-ZandraFan.json")[0]
    add_event = {
        "type": "add",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    # Simulate a poll
    bridge.events.emit(event.EventType.RESOURCE_ADDED, add_event)
    # Bad way to check, but just wait a second so it can get processed
    await asyncio.sleep(1)
    assert len(bridge.fans._items) == 1
    # Simulate an update
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="available",
            functionInstance=None,
            value=False,
        ),
    )
    update_event = {
        "type": "update",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    bridge.events.emit(event.EventType.RESOURCE_UPDATED, update_event)
    # Bad way to check, but just wait a second so it can get processed
    await asyncio.sleep(1)
    assert len(bridge.fans._items) == 1
    assert not bridge.fans._items[dev_update.id].available



from dataclasses import dataclass, field

from ..models import features
from .resource import DeviceInformation, ResourceTypes
from .sensor import AferoBinarySensor, AferoSensor


@dataclass
class Thermostat:
    """Representation of an Afero Thermostat"""

    id: str  # ID used when interacting with Afero
    available: bool

    current_temperature: float | None
    fan_running: bool | None
    fan_mode: features.ModeFeature | None
    hvac_action: str | None
    hvac_mode: features.HVACModeFeature | None
    safety_max_temp: features.TargetTemperatureFeature | None
    safety_min_temp: features.TargetTemperatureFeature | None
    target_temperature_auto_heating: features.TargetTemperatureFeature | None
    target_temperature_auto_cooling: features.TargetTemperatureFeature | None
    target_temperature_heating: features.TargetTemperatureFeature | None
    target_temperature_cooling: features.TargetTemperatureFeature | None

    # Defined at initialization
    instances: dict = field(default_factory=lambda: dict(), repr=False, init=False)
    device_information: DeviceInformation = field(default_factory=DeviceInformation)
    sensors: dict[str, AferoSensor] = field(default_factory=lambda: dict())
    binary_sensors: dict[str, AferoBinarySensor] = field(default_factory=lambda: dict())

    type: ResourceTypes = ResourceTypes.THERMOSTAT

    def __init__(self, functions: list, **kwargs):
        for key, value in kwargs.items():
            if key == "instances":
                continue
            setattr(self, key, value)
        instances = {}
        for function in functions:
            instances[function["functionClass"]] = function.get(
                "functionInstance", None
            )
        self.instances = instances

    @property
    def target_temperature(self) -> float | None:
        if self.hvac_mode is None or self.hvac_mode.mode not in [
            "cool",
            "heat",
            "fan",
            "off",
        ]:
            return None
        return getattr(
            self._get_target_feature(self.get_mode_to_check()), "value", None
        )

    def get_mode_to_check(self) -> str | None:
        if not self.hvac_mode:
            return None
        if self.hvac_mode.mode in ["cool", "heat"]:
            return self.hvac_mode.mode
        elif self.hvac_mode.previous_mode in ["cool", "heat"]:
            return self.hvac_mode.previous_mode
        else:
            return None

    def _get_target_feature(self, mode: str) -> float | None:
        if mode == "cool":
            return self.target_temperature_cooling
        elif mode == "heat":
            return self.target_temperature_heating
        else:
            return None

    @property
    def target_temperature_range(self) -> tuple[float, float]:
        return (
            self.target_temperature_auto_heating.value,
            self.target_temperature_auto_cooling.value,
        )

    @property
    def target_temperature_step(self) -> float:
        set_mode = self.get_mode_to_check()
        if not set_mode:
            return 0.5  # Default from Hubspace
        else:
            return getattr(
                self._get_target_feature(self.get_mode_to_check()), "step", None
            )

    @property
    def target_temperature_max(self) -> float:
        set_mode = self.get_mode_to_check()
        if not set_mode or self.hvac_mode.mode == "auto":
            return self.target_temperature_auto_cooling.max
        else:
            return getattr(self._get_target_feature(set_mode), "max", None)

    @property
    def target_temperature_min(self) -> float | None:
        set_mode = self.get_mode_to_check()
        if not set_mode or self.hvac_mode.mode == "auto":
            return self.target_temperature_auto_heating.min
        else:
            return getattr(
                self._get_target_feature(self.get_mode_to_check()), "min", None
            )

    @property
    def supports_fan_mode(self) -> bool:
        return self.fan_mode is not None

    @property
    def supports_temperature_range(self) -> bool:
        return (
            self.target_temperature_auto_cooling is not None
            and self.target_temperature_auto_heating is not None
        )

    def get_instance(self, elem):
        """Lookup the instance associated with the elem"""
        return self.instances.get(elem, None)


@dataclass
class ThermostatPut:
    """States that can be updated for a Thermostat"""

    fan_mode: features.ModeFeature | None = None
    hvac_mode: features.HVACModeFeature | None = None
    safety_max_temp: features.TargetTemperatureFeature | None = None
    safety_min_temp: features.TargetTemperatureFeature | None = None
    target_temperature_auto_heating: features.TargetTemperatureFeature | None = None
    target_temperature_auto_cooling: features.TargetTemperatureFeature | None = None
    target_temperature_heating: features.TargetTemperatureFeature | None = None
    target_temperature_cooling: features.TargetTemperatureFeature | None = None



"""Controller holding and managing Afero IoT resources of type `lock`."""

from ...device import AferoDevice
from ..models import features
from ..models.lock import Lock, LockPut
from ..models.resource import DeviceInformation, ResourceTypes
from .base import AferoBinarySensor, AferoSensor, BaseResourcesController


class LockController(BaseResourcesController[Lock]):
    """Controller holding and managing Afero IoT resources of type `lock`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.LOCK]
    ITEM_CLS = Lock
    ITEM_MAPPING = {"position": "lock-control"}
    # Sensors map functionClass -> Unit
    ITEM_SENSORS: dict[str, str] = {}
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {}

    async def lock(self, device_id: str) -> None:
        """Engage the lock"""
        await self.set_state(
            device_id, lock_position=features.CurrentPositionEnum.LOCKING
        )

    async def unlock(self, device_id: str) -> None:
        """Disengage the lock"""
        await self.set_state(
            device_id, lock_position=features.CurrentPositionEnum.UNLOCKING
        )

    async def initialize_elem(self, afero_device: AferoDevice) -> Lock:
        """Initialize the element"""
        available: bool = False
        current_position: features.CurrentPositionFeature | None = None
        sensors: dict[str, AferoSensor] = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        for state in afero_device.states:
            if state.functionClass == "lock-control":
                current_position = features.CurrentPositionFeature(
                    position=features.CurrentPositionEnum(state.value)
                )
            elif state.functionClass == "available":
                available = state.value
            elif sensor := await self.initialize_sensor(state, afero_device.device_id):
                if isinstance(sensor, AferoBinarySensor):
                    binary_sensors[sensor.id] = sensor
                else:
                    sensors[sensor.id] = sensor

        self._items[afero_device.id] = Lock(
            afero_device.functions,
            id=afero_device.id,
            available=available,
            sensors=sensors,
            binary_sensors=binary_sensors,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
            ),
            position=current_position,
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        cur_item = self.get_device(afero_device.id)
        updated_keys = set()
        for state in afero_device.states:
            if state.functionClass == "lock-control":
                new_val = features.CurrentPositionEnum(state.value)
                if cur_item.position.position != new_val:
                    updated_keys.add("position")
                cur_item.position.position = features.CurrentPositionEnum(state.value)
            elif state.functionClass == "available":
                if cur_item.available != state.value:
                    updated_keys.add("available")
                cur_item.available = state.value
            elif update_key := await self.update_sensor(state, cur_item):
                updated_keys.add(update_key)

        return updated_keys

    async def set_state(
        self,
        device_id: str,
        lock_position: features.CurrentPositionEnum | None = None,
    ) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = LockPut()
        if lock_position is not None:
            update_obj.position = features.CurrentPositionFeature(
                position=lock_position
            )
        await self.update(device_id, obj_in=update_obj)



from dataclasses import dataclass, field

from .resource import DeviceInformation, ResourceTypes
from .sensor import AferoBinarySensor, AferoSensor


@dataclass
class Device:
    """Representation of an Afero parent item"""

    id: str  # ID used when interacting with Afero
    available: bool

    device_information: DeviceInformation = field(default_factory=DeviceInformation)
    sensors: dict[str, AferoSensor] = field(default_factory=lambda: dict())
    binary_sensors: dict[str, AferoBinarySensor] = field(default_factory=lambda: dict())

    type: ResourceTypes = ResourceTypes.PARENT_DEVICE



import json
import os
from typing import Any

from aioafero.device import AferoDevice, AferoState

current_path: str = os.path.dirname(os.path.realpath(__file__))


def get_device_dump(file_name: str) -> Any:
    """Get a device dump

    :param file_name: Name of the file to load
    """
    with open(os.path.join(current_path, "device_dumps", file_name), "r") as fh:
        return json.load(fh)


def get_raw_dump(file_name: str) -> Any:
    """Get a device dump

    :param file_name: Name of the file to load
    """
    with open(os.path.join(current_path, "data", file_name), "r") as fh:
        return json.load(fh)


def create_devices_from_data(file_name: str) -> list[AferoDevice]:
    """Generate devices from a data dump

    :param file_name: Name of the file to load
    """
    devices = get_device_dump(file_name)
    processed = []
    for device in devices:
        processed.append(create_device_from_data(device))
    return processed


def create_device_from_data(device: dict) -> AferoDevice:
    processed_states = []
    for state in device["states"]:
        processed_states.append(AferoState(**state))
    device["states"] = processed_states
    if "children" not in device:
        device["children"] = []
    return AferoDevice(**device)


def get_json_call(mocked_controller):
    mocked_controller._bridge.request.assert_called_once()
    call = mocked_controller._bridge.request.call_args_list[0][1]
    assert "json" in call
    return call["json"]


def ensure_states_sent(mocked_controller, expected_states):
    req = get_json_call(mocked_controller)["values"]
    assert len(req) == len(
        expected_states
    ), f"States Sent: {len(req)}. Expected: {len(expected_states)}. Actual: {req}"
    for state in expected_states:
        assert state in req, (
            f"Missing {state['functionClass']} / "
            f"{state['functionInstance']} for "
            f"{state['value']} in {req}"
        )


def modify_state(device: AferoDevice, new_state):
    for ind, state in enumerate(device.states):
        if state.functionClass != new_state.functionClass:
            continue
        if (
            new_state.functionInstance
            and new_state.functionInstance != state.functionInstance
        ):
            continue
        device.states[ind] = new_state
        break



import datetime

import pytest
import pytest_asyncio
from aioresponses import aioresponses

from aioafero.v1 import AferoBridgeV1
from aioafero.v1.auth import token_data
from aioafero.v1.controllers.event import EventType


@pytest.fixture
def mocked_bridge(mocker):
    bridge: AferoBridgeV1 = AferoBridgeV1("username2", "password2")
    mocker.patch.object(
        bridge,
        "get_account_id",
        side_effect=mocker.AsyncMock(return_value="mocked-account-id"),
    )
    mocker.patch.object(bridge, "_account_id", "mocked-account-id")
    mocker.patch.object(bridge, "fetch_data", return_value=[])
    mocker.patch.object(bridge, "request", side_effect=mocker.AsyncMock())
    mocker.patch.object(bridge, "initialize", side_effect=mocker.AsyncMock())
    # Force initialization so test elements are not overwritten
    for controller in bridge._controllers:
        controller._initialized = True

    # Enable ad-hoc event updates
    def emit_event(event_type, data):
        bridge.events.emit(EventType(event_type), data)

    bridge.emit_event = emit_event
    bridge.__aenter__ = mocker.AsyncMock(return_value=bridge)
    bridge.__aexit__ = mocker.AsyncMock()
    yield bridge


@pytest.fixture
def mocked_bridge_req(mocker):
    bridge: AferoBridgeV1 = AferoBridgeV1("username2", "password2")
    mocker.patch.object(
        bridge,
        "get_account_id",
        side_effect=mocker.AsyncMock(return_value="mocked-account-id"),
    )
    mocker.patch.object(bridge, "_account_id", "mocked-account-id")
    mocker.patch.object(bridge, "initialize", side_effect=mocker.AsyncMock())
    mocker.patch.object(bridge, "fetch_data", side_effect=bridge.fetch_data)
    mocker.patch.object(bridge, "request", side_effect=bridge.request)
    bridge._auth._token_data = token_data(
        "mock-token",
        None,
        "mock-refresh-token",
        expiration=datetime.datetime.now().timestamp() + 200,
    )
    # Force initialization so test elements are not overwritten
    for controller in bridge._controllers:
        controller._initialized = True

    # Enable ad-hoc event updates
    def emit_event(event_type, data):
        bridge.events.emit(EventType(event_type), data)

    bridge.emit_event = emit_event
    bridge.__aenter__ = mocker.AsyncMock(return_value=bridge)
    bridge.__aexit__ = mocker.AsyncMock()
    yield bridge


@pytest_asyncio.fixture
async def bridge(mocker):
    bridge = AferoBridgeV1("user", "passwd")
    mocker.patch.object(bridge, "_account_id", "mocked-account-id")
    mocker.patch.object(bridge, "fetch_data", return_value=[])
    mocker.patch.object(bridge, "request", side_effect=mocker.AsyncMock())
    await bridge.initialize()
    yield bridge
    await bridge.close()


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m



"""Test ExhaustFanController"""

import asyncio

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers import event
from aioafero.v1.controllers.exhaust_fan import ExhaustFanController, features
from aioafero.v1.models.sensor import AferoBinarySensor

from .. import utils

exhaust_fan = utils.create_devices_from_data("exhaust-fan.json")[0]


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    controller = ExhaustFanController(mocked_bridge)
    yield controller


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller.initialize_elem(exhaust_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == "44620d02-8b62-49ce-afe8-1ea8f15e0ec5"
    assert dev.binary_sensors == {
        "humidity-threshold-met|humidity-threshold-met": AferoBinarySensor(
            id="humidity-threshold-met|humidity-threshold-met",
            owner="7475607b-c3ea-4afe-a465-1c11c5c09985",
            _value="below-threshold",
            _error="above-threshold",
            unit=None,
            instance="humidity-threshold-met",
        ),
        "motion-detection|motion-detection": AferoBinarySensor(
            id="motion-detection|motion-detection",
            owner="7475607b-c3ea-4afe-a465-1c11c5c09985",
            _value="motion-detected",
            _error="motion-detected",
            unit=None,
            instance="motion-detection",
        ),
    }
    assert dev.sensors == {}
    assert dev.numbers == {
        ("auto-off-timer", "auto-off"): features.NumbersFeature(
            value=1200,
            min=60,
            max=1800,
            step=60,
            name="Auto Off Timer",
            unit="seconds",
        )
    }
    assert dev.selects == {
        ("motion-action", "exhaust-fan"): features.SelectFeature(
            selected="light", selects={"both", "fan", "light"}, name="Motion Action"
        ),
        ("sensitivity", "humidity-sensitivity"): features.SelectFeature(
            selected="3-medium",
            selects={"1-low", "2-low-medium", "3-medium", "4-medium-high", "5-high"},
            name="Humidity Sensitivity",
        ),
    }


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller.initialize_elem(exhaust_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.available is True
    dev_update = utils.create_devices_from_data("exhaust-fan.json")[0]
    new_states = [
        AferoState(
            **{
                "functionClass": "available",
                "value": False,
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "humidity-threshold-met",
                "value": "above-threshold",
                "lastUpdateTime": 0,
                "functionInstance": "humidity-threshold-met",
            }
        ),
        AferoState(
            **{
                "functionClass": "auto-off-timer",
                "value": 120,
                "lastUpdateTime": 0,
                "functionInstance": "auto-off",
            }
        ),
        AferoState(
            **{
                "functionClass": "motion-action",
                "value": "both",
                "lastUpdateTime": 0,
                "functionInstance": "exhaust-fan",
            }
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.available is False
    assert (
        dev.binary_sensors["humidity-threshold-met|humidity-threshold-met"]._value
        == "above-threshold"
    )
    assert dev.numbers[("auto-off-timer", "auto-off")].value == 120
    assert dev.selects[("motion-action", "exhaust-fan")].selected == "both"
    assert updates == {
        "binary-humidity-threshold-met|humidity-threshold-met",
        "available",
        "number-('auto-off-timer', 'auto-off')",
        "select-('motion-action', 'exhaust-fan')",
    }


@pytest.mark.asyncio
async def test_update_state_no_change(mocked_controller):
    await mocked_controller.initialize_elem(exhaust_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.available is True
    dev_update = utils.create_devices_from_data("exhaust-fan.json")[0]
    new_states = [
        AferoState(
            **{
                "functionClass": "available",
                "value": True,
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.available is True
    assert updates == set()


@pytest.mark.asyncio
async def test_set_state_empty(mocked_controller):
    await mocked_controller.initialize_elem(exhaust_fan)
    await mocked_controller.set_state(exhaust_fan.id)


@pytest.mark.asyncio
async def test_set_state(mocked_controller):
    await mocked_controller.initialize_elem(exhaust_fan)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.available is True
    await mocked_controller.set_state(
        exhaust_fan.id,
        numbers={
            ("auto-off-timer", "auto-off"): 120,
            ("invalid", "state"): None,
        },
        selects={
            ("motion-action", "exhaust-fan"): "both",
            ("invalid", "state"): None,
        },
    )
    dev = mocked_controller.items[0]
    assert dev.numbers[("auto-off-timer", "auto-off")].value == 120
    assert dev.selects[("motion-action", "exhaust-fan")].selected == "both"


@pytest.mark.asyncio
async def test_exhaust_fan_emitting(bridge):
    dev_update = utils.create_devices_from_data("exhaust-fan.json")[0]
    add_event = {
        "type": "add",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    # Simulate a poll
    bridge.events.emit(event.EventType.RESOURCE_ADDED, add_event)
    # Bad way to check, but just wait a second so it can get processed
    await asyncio.sleep(1)
    assert len(bridge.exhaust_fans._items) == 1
    # Simulate an update
    utils.modify_state(
        dev_update,
        AferoState(
            functionClass="available",
            functionInstance=None,
            value=False,
        ),
    )
    update_event = {
        "type": "update",
        "device_id": dev_update.id,
        "device": dev_update,
    }
    bridge.events.emit(event.EventType.RESOURCE_UPDATED, update_event)
    # Bad way to check, but just wait a second so it can get processed
    await asyncio.sleep(1)
    assert len(bridge.exhaust_fans._items) == 1
    assert not bridge.exhaust_fans._items[dev_update.id].available


@pytest.mark.asyncio
async def test_set_state_no_dev(mocked_controller, caplog):
    caplog.set_level(0)
    await mocked_controller.initialize_elem(exhaust_fan)
    mocked_controller._bridge.add_device(exhaust_fan.id, mocked_controller)
    await mocked_controller.set_state("not-a-device")
    mocked_controller._bridge.request.assert_not_called()
    assert "Unable to find device" in caplog.text



"""Controller holding and managing Afero IoT resources of type `thermostat`."""

import copy

from ... import device
from ...device import AferoDevice, AferoState
from ...util import process_function
from ..models import features
from ..models.resource import DeviceInformation, ResourceTypes
from ..models.thermostat import Thermostat, ThermostatPut
from .base import AferoBinarySensor, AferoSensor, BaseResourcesController


class ThermostatController(BaseResourcesController[Thermostat]):
    """Controller holding and managing Afero IoT resources of type `thermostat`."""

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.THERMOSTAT]
    ITEM_CLS = Thermostat
    ITEM_MAPPING = {
        "fan_mode": "fan-mode",
        "hvac_mode": "mode",
    }
    # Sensors map functionClass -> Unit
    ITEM_SENSORS: dict[str, str] = {}
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {
        "filter-replacement": "replacement-needed",
        "max-temp-exceeded": "alerting",
        "min-temp-exceeded": "alerting",
    }

    async def initialize_elem(self, afero_device: AferoDevice) -> Thermostat:
        """Initialize the element"""
        available: bool = False
        current_temperature: float | None = None
        fan_running: bool | None = None
        fan_mode: features.ModeFeature | None = None
        hvac_action: str | None = None
        hvac_mode: features.HVACModeFeature | None = None
        safety_max_temp: features.TargetTemperatureFeature | None = None
        safety_min_temp: features.TargetTemperatureFeature | None = None
        target_temperature_auto_heating: features.TargetTemperatureFeature | None = None
        target_temperature_auto_cooling: features.TargetTemperatureFeature | None = None
        target_temperature_heating: features.TargetTemperatureFeature | None = None
        target_temperature_cooling: features.TargetTemperatureFeature | None = None
        sensors: dict[str, AferoSensor] = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        all_modes: set[str] | None = None
        current_mode: str | None = None
        system_type: str | None = None
        for state in afero_device.states:
            func_def = device.get_function_from_device(
                afero_device.functions, state.functionClass, state.functionInstance
            )
            if state.functionClass == "fan-mode":
                fan_mode = features.ModeFeature(
                    mode=state.value,
                    modes=set(process_function(afero_device.functions, "fan-mode")),
                )
            if state.functionClass == "current-fan-state":
                fan_running = state.value == "on"
            elif state.functionClass == "temperature":
                if state.functionInstance == "current-temp":
                    current_temperature = round(state.value, 2)
                elif state.functionInstance == "safety-mode-max-temp":
                    safety_max_temp = generate_target_temp(func_def["values"][0], state)
                elif state.functionInstance == "safety-mode-min-temp":
                    safety_min_temp = generate_target_temp(func_def["values"][0], state)
                elif state.functionInstance == "auto-heating-target":
                    target_temperature_auto_heating = generate_target_temp(
                        func_def["values"][0], state
                    )
                elif state.functionInstance == "auto-cooling-target":
                    target_temperature_auto_cooling = generate_target_temp(
                        func_def["values"][0], state
                    )
                elif state.functionInstance == "heating-target":
                    target_temperature_heating = generate_target_temp(
                        func_def["values"][0], state
                    )
                elif state.functionInstance == "cooling-target":
                    target_temperature_cooling = generate_target_temp(
                        func_def["values"][0], state
                    )
            elif state.functionClass == "mode":
                all_modes = set(process_function(afero_device.functions, "mode"))
                current_mode = state.value
            elif state.functionClass == "current-system-state":
                hvac_action = state.value
            elif state.functionClass == "system-type":
                system_type = state.value
            elif state.functionClass == "available":
                available = state.value
            elif sensor := await self.initialize_sensor(state, afero_device.id):
                if isinstance(sensor, AferoBinarySensor):
                    binary_sensors[sensor.id] = sensor
                else:
                    sensors[sensor.id] = sensor

        # Determine supported modes
        if current_mode and all_modes and system_type:
            hvac_mode = features.HVACModeFeature(
                mode=current_mode,
                previous_mode=current_mode,
                modes=all_modes,
                supported_modes=get_supported_modes(system_type, all_modes),
            )

        self._items[afero_device.id] = Thermostat(
            afero_device.functions,
            id=afero_device.id,
            available=available,
            sensors=sensors,
            binary_sensors=binary_sensors,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
            ),
            current_temperature=current_temperature,
            fan_running=fan_running,
            fan_mode=fan_mode,
            hvac_action=hvac_action,
            hvac_mode=hvac_mode,
            safety_max_temp=safety_max_temp,
            safety_min_temp=safety_min_temp,
            target_temperature_auto_cooling=target_temperature_auto_cooling,
            target_temperature_auto_heating=target_temperature_auto_heating,
            target_temperature_cooling=target_temperature_cooling,
            target_temperature_heating=target_temperature_heating,
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        updated_keys = set()
        cur_item = self.get_device(afero_device.id)
        temperature_mapping = {
            "auto-heating-target": "target_temperature_auto_heating",
            "auto-cooling-target": "target_temperature_auto_cooling",
            "cooling-target": "target_temperature_cooling",
            "heating-target": "target_temperature_heating",
            "safety-mode-max-temp": "safety_max_temp",
            "safety-mode-min-temp": "safety_min_temp",
        }
        temp_mapping_raw = {
            "current-temp": "current_temperature",
        }
        for state in afero_device.states:
            if state.functionClass == "current-fan-state":
                temp_val = state.value == "on"
                if cur_item.fan_running != temp_val:
                    cur_item.fan_running = temp_val
                    updated_keys.add("current-fan-state")
            elif state.functionClass == "fan-mode":
                if cur_item.fan_mode.mode != state.value:
                    cur_item.fan_mode.mode = state.value
                    updated_keys.add("fan-mode")
            elif state.functionClass == "mode":
                if cur_item.hvac_mode.mode != state.value:
                    # We only want to update the previous mode when we are in heat or cool
                    if cur_item.hvac_mode.mode in ["cool", "heat"]:
                        cur_item.hvac_mode.previous_mode = cur_item.hvac_mode.mode
                    cur_item.hvac_mode.mode = state.value
                    updated_keys.add(state.functionClass)
            elif state.functionClass == "temperature":
                rounded_value = round(state.value, 2)
                if state.functionInstance in temp_mapping_raw.keys():
                    temp_value = getattr(
                        cur_item, temp_mapping_raw.get(state.functionInstance), None
                    )
                    if temp_value != rounded_value:
                        setattr(
                            cur_item,
                            temp_mapping_raw.get(state.functionInstance),
                            rounded_value,
                        )
                        updated_keys.add(f"temperature-{state.functionInstance}")
                elif state.functionInstance in temperature_mapping.keys():
                    temp_item = getattr(
                        cur_item, temperature_mapping.get(state.functionInstance), None
                    )
                    if temp_item.value != rounded_value:
                        temp_item.value = rounded_value
                        updated_keys.add(f"temperature-{state.functionInstance}")
            elif state.functionClass == "current-system-state":
                if cur_item.hvac_action != state.value:
                    cur_item.hvac_action = state.value
                    updated_keys.add(state.functionClass)
            elif state.functionClass == "available":
                if cur_item.available != state.value:
                    cur_item.available = state.value
                    updated_keys.add("available")
            elif update_key := await self.update_sensor(state, cur_item):
                updated_keys.add(update_key)
        return updated_keys

    async def set_fan_mode(self, device_id: str, fan_mode: str) -> None:
        return await self.set_state(device_id, fan_mode=fan_mode)

    async def set_hvac_mode(self, device_id: str, hvac_mode: str) -> None:
        return await self.set_state(device_id, hvac_mode=hvac_mode)

    async def set_target_temperature(
        self, device_id: str, target_temperature: float
    ) -> None:
        return await self.set_state(device_id, target_temperature=target_temperature)

    async def set_temperature_range(
        self, device_id: str, temp_low: float, temp_high: float
    ) -> None:
        return await self.set_state(
            device_id,
            target_temperature_auto_heating=temp_low,
            target_temperature_auto_cooling=temp_high,
        )

    async def set_state(
        self,
        device_id: str,
        fan_mode: str | None = None,
        hvac_mode: str | None = None,
        safety_max_temp: float | None = None,
        safety_min_temp: float | None = None,
        target_temperature_auto_heating: float | None = None,
        target_temperature_auto_cooling: float | None = None,
        target_temperature_heating: float | None = None,
        target_temperature_cooling: float | None = None,
        target_temperature: float | None = None,
    ) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = ThermostatPut()
        cur_item = self.get_device(device_id)
        if fan_mode is not None:
            if fan_mode in cur_item.fan_mode.modes:
                update_obj.fan_mode = features.ModeFeature(
                    mode=fan_mode,
                    modes=cur_item.fan_mode.modes,
                )
                update_obj.hvac_mode = features.HVACModeFeature(
                    mode="fan",
                    modes=cur_item.hvac_mode.modes,
                    previous_mode=cur_item.hvac_mode.mode,
                    supported_modes=cur_item.hvac_mode.supported_modes,
                )
            else:
                self._logger.debug(
                    "Unknown fan mode %s. Available modes: %s",
                    fan_mode,
                    ", ".join(sorted(cur_item.fan_mode.modes)),
                )
        if hvac_mode is not None and not update_obj.hvac_mode:
            if hvac_mode in cur_item.hvac_mode.supported_modes:
                update_obj.hvac_mode = features.HVACModeFeature(
                    mode=hvac_mode,
                    modes=cur_item.hvac_mode.modes,
                    previous_mode=cur_item.hvac_mode.mode,
                    supported_modes=cur_item.hvac_mode.supported_modes,
                )
            else:
                self._logger.debug(
                    "Unknown hvac mode %s. Available modes: %s",
                    hvac_mode,
                    ", ".join(sorted(cur_item.hvac_mode.modes)),
                )
        # Setting the temp without a specific means we need to adjust the active
        # mode.
        if target_temperature:
            if hvac_mode and hvac_mode in cur_item.hvac_mode.supported_modes:
                mode_to_set = hvac_mode
            else:
                mode_to_set = cur_item.get_mode_to_check()
            if mode_to_set == "cool":
                target_temperature_cooling = target_temperature
            elif mode_to_set == "heat":
                target_temperature_heating = target_temperature
            else:
                self._logger.debug(
                    "Unable to set the target temperature due to the active mode: %s",
                    cur_item.hvac_mode.mode,
                )
        if safety_min_temp is not None:
            update_obj.safety_min_temp = features.TargetTemperatureFeature(
                value=safety_min_temp,
                min=cur_item.safety_min_temp.min,
                max=cur_item.safety_min_temp.max,
                step=cur_item.safety_min_temp.step,
                instance=cur_item.safety_min_temp.instance,
            )
        if safety_max_temp is not None:
            update_obj.safety_max_temp = features.TargetTemperatureFeature(
                value=safety_max_temp,
                min=cur_item.safety_max_temp.min,
                max=cur_item.safety_max_temp.max,
                step=cur_item.safety_max_temp.step,
                instance=cur_item.safety_max_temp.instance,
            )
        if target_temperature_auto_heating is not None:
            update_obj.target_temperature_auto_heating = (
                features.TargetTemperatureFeature(
                    value=target_temperature_auto_heating,
                    min=cur_item.target_temperature_auto_heating.min,
                    max=cur_item.target_temperature_auto_heating.max,
                    step=cur_item.target_temperature_auto_heating.step,
                    instance=cur_item.target_temperature_auto_heating.instance,
                )
            )
        if target_temperature_auto_cooling is not None:
            update_obj.target_temperature_auto_cooling = (
                features.TargetTemperatureFeature(
                    value=target_temperature_auto_cooling,
                    min=cur_item.target_temperature_auto_cooling.min,
                    max=cur_item.target_temperature_auto_cooling.max,
                    step=cur_item.target_temperature_auto_cooling.step,
                    instance=cur_item.target_temperature_auto_cooling.instance,
                )
            )
        if target_temperature_heating is not None:
            update_obj.target_temperature_heating = features.TargetTemperatureFeature(
                value=target_temperature_heating,
                min=cur_item.target_temperature_heating.min,
                max=cur_item.target_temperature_heating.max,
                step=cur_item.target_temperature_heating.step,
                instance=cur_item.target_temperature_heating.instance,
            )
        if target_temperature_cooling is not None:
            update_obj.target_temperature_cooling = features.TargetTemperatureFeature(
                value=target_temperature_cooling,
                min=cur_item.target_temperature_cooling.min,
                max=cur_item.target_temperature_cooling.max,
                step=cur_item.target_temperature_cooling.step,
                instance=cur_item.target_temperature_cooling.instance,
            )
        await self.update(device_id, obj_in=update_obj)


def generate_target_temp(
    func_def: dict, state: AferoState
) -> features.TargetTemperatureFeature:
    return features.TargetTemperatureFeature(
        value=round(state.value, 2),
        step=func_def["range"]["step"],
        min=func_def["range"]["min"],
        max=func_def["range"]["max"],
        instance=state.functionInstance,
    )


def get_supported_modes(system_type: str, all_modes: set[str]) -> set:
    """Determine the supported modes based on the system_type."""
    supported_modes = copy.copy(all_modes)
    if "heat-pump" in system_type:
        supports_heating = True
        supports_cooling = True
    else:
        supports_heating = "heating" in system_type
        supports_cooling = "cooling" in system_type
    if not supports_heating and "heat" in supported_modes:
        supported_modes.remove("heat")
    if not supports_cooling and "cool" in supported_modes:
        supported_modes.remove("cool")
    if (not supports_cooling or not supports_heating) and "auto" in supported_modes:
        supported_modes.remove("auto")
    return supported_modes



"""Handle connecting to Afero IoT and distribute events."""

import asyncio
from asyncio.coroutines import iscoroutinefunction
from collections.abc import Callable
from enum import Enum
from types import NoneType
from typing import TYPE_CHECKING, Any, NotRequired, TypedDict

from aiohttp.client_exceptions import ClientError
from aiohttp.web_exceptions import HTTPForbidden, HTTPTooManyRequests

from ...device import AferoDevice, get_afero_device
from ...errors import InvalidAuth
from ...types import EventType

if TYPE_CHECKING:  # pragma: no cover
    from .. import AferoBridgeV1


class BackoffException(Exception):
    """Exception raised when a backoff is required."""

    pass


class EventStreamStatus(Enum):
    """Status options of EventStream."""

    CONNECTING = 0
    CONNECTED = 1
    DISCONNECTED = 2


class AferoEvent(TypedDict):
    """Afero IoT Event message as emitted by the EventStream."""

    type: EventType  # = EventType (add, update, delete)
    device_id: NotRequired[str]  # ID for interacting with the device
    device: NotRequired[AferoDevice]  # Afero Device
    force_forward: NotRequired[bool]


EventCallBackType = Callable[[EventType, dict | None], None]
EventSubscriptionType = tuple[
    EventCallBackType,
    "tuple[EventType] | None",
    "tuple[ResourceTypes] | None",
]


class EventStream:

    def __init__(self, bridge: "AferoBridgeV1", polling_interval: int) -> None:
        """Initialize instance."""
        self._bridge = bridge
        self._listeners = set()
        self._event_queue = asyncio.Queue()
        self._status = EventStreamStatus.DISCONNECTED
        self._bg_tasks: list[asyncio.Task] = []
        self._subscribers: list[EventSubscriptionType] = []
        self._logger = bridge.logger.getChild("events")
        self._polling_interval: int = polling_interval

    @property
    def connected(self) -> bool:
        """Return bool if we're connected."""
        return self._status == EventStreamStatus.CONNECTED

    @property
    def status(self) -> EventStreamStatus:
        """Return connection status."""
        return self._status

    @property
    def polling_interval(self) -> int:
        return self._polling_interval

    @polling_interval.setter
    def polling_interval(self, polling_interval: int) -> None:
        self._polling_interval = polling_interval

    async def initialize(self) -> None:
        """Start the polling processes"""
        assert len(self._bg_tasks) == 0
        await self.initialize_reader()
        await self.initialize_processor()

    async def initialize_reader(self) -> None:
        self._bg_tasks.append(asyncio.create_task(self.__event_reader()))

    async def initialize_processor(self) -> None:
        self._bg_tasks.append(asyncio.create_task(self.__event_processor()))

    async def stop(self) -> None:
        """Stop listening for events."""
        for task in self._bg_tasks:
            task.cancel()
        self._status = EventStreamStatus.DISCONNECTED
        self._bg_tasks = []

    def subscribe(
        self,
        callback: EventCallBackType,
        event_filter: EventType | tuple[EventType] | None = None,
        resource_filter: tuple[str] | None = None,
    ) -> Callable:
        """
        Subscribe to events emitted

        Parameters:
            - `callback` - callback function to call when an event emits.
            - `event_filter` - Optionally provide an EventType as filter.
            - `resource_filter` - Optionally provide a ResourceType as filter.

        Returns:
            function to unsubscribe.
        """
        if not isinstance(event_filter, NoneType | tuple):
            event_filter = (event_filter,)
        if not isinstance(resource_filter, NoneType | tuple):
            resource_filter = (resource_filter,)
        subscription = (callback, event_filter, resource_filter)

        def unsubscribe():
            self._subscribers.remove(subscription)

        self._subscribers.append(subscription)
        return unsubscribe

    def add_job(self, event: AferoEvent) -> None:
        """Manually add a job to be processed."""
        self._event_queue.put_nowait(event)

    def emit(self, event_type: EventType, data: AferoEvent = None) -> None:
        """Emit event to all listeners."""
        for callback, event_filter, resource_filter in self._subscribers:
            try:
                if event_filter is not None and event_type not in event_filter:
                    continue
                if (
                    resource_filter is not None
                    and data is not None
                    and (
                        "device" in data
                        and data["device"]
                        and not any(
                            data["device"].device_class == res_filter
                            for res_filter in resource_filter
                        )
                    )
                ):
                    continue
                if iscoroutinefunction(callback):
                    asyncio.create_task(callback(event_type, data))
                else:
                    callback(event_type, data)
            except Exception:
                self._logger.exception("Unhandled exception. Please open a bug report")

    async def process_backoff(self, attempt: int) -> None:
        """Handle backoff timer for Afero IoT API

        :param attempt: Number of attempts
        :param reason: Reason why the backoff is occurring
        """
        backoff_time = min(attempt * self.polling_interval, 600)
        debug_message = f"Waiting {backoff_time} seconds before next poll"
        if attempt == 1:
            self._logger.info("Lost connection to the Afero IoT API.")
            self._logger.debug(debug_message)
        if self._status != EventStreamStatus.DISCONNECTED:
            self._status = EventStreamStatus.DISCONNECTED
            self.emit(EventType.DISCONNECTED)
        await asyncio.sleep(backoff_time)

    async def gather_data(self) -> list[dict[Any, str]]:
        """Gather all data from the Afero IoT API"""
        consecutive_http_errors = 0
        while True:
            try:
                data = await self._bridge.fetch_data()
            except asyncio.TimeoutError:
                self._logger.warning("Timeout when contacting Afero IoT API.")
                await self.process_backoff(consecutive_http_errors)
            except InvalidAuth:
                consecutive_http_errors += 1
                self._logger.warning("Invalid credentials provided.")
                await self.process_backoff(consecutive_http_errors)
            except (HTTPForbidden, HTTPTooManyRequests, ClientError):
                consecutive_http_errors += 1
                await self.process_backoff(consecutive_http_errors)
            except ValueError as err:
                self._logger.warning(
                    "Unexpected data from Afero IoT API, %s.", err.args[0]
                )
                consecutive_http_errors += 1
                await self.process_backoff(consecutive_http_errors)
            except Exception as err:
                self._logger.exception(
                    "Unknown error occurred. Please open a bug report."
                )
                raise err
            else:
                # Successful connection
                if consecutive_http_errors > 0:
                    self._logger.info("Reconnected to the Afero IoT API")
                    self.emit(EventType.RECONNECTED)
                elif self._status != EventStreamStatus.CONNECTED:
                    self._status = EventStreamStatus.CONNECTED
                    self.emit(EventType.CONNECTED)
                return data

    async def generate_events_from_data(self, data: list[dict[Any, str]]) -> None:
        """Process the raw Afero IoT data for emitting

        :param data: Raw data from Afero IoT
        """
        processed_ids = []
        skipped_ids = []
        for dev in data:
            device = get_afero_device(dev)
            if not device.device_class:
                continue
            event_type = EventType.RESOURCE_UPDATED
            if device.id not in self._bridge.tracked_devices:
                event_type = EventType.RESOURCE_ADDED
            self._event_queue.put_nowait(
                AferoEvent(
                    type=event_type,
                    device_id=device.id,
                    device=device,
                    force_forward=False,
                )
            )
            processed_ids.append(device.id)
        # Handle devices that did not report in from the API
        for dev_id in self._bridge.tracked_devices:
            if dev_id not in processed_ids + skipped_ids:
                self._event_queue.put_nowait(
                    AferoEvent(type=EventType.RESOURCE_DELETED, device_id=dev_id)
                )
                self._bridge.remove_device(dev_id)

    async def perform_poll(self) -> None:
        """Poll Afero IoT and generate the required events"""
        try:
            data = await self.gather_data()
        except Exception:
            self._status = EventStreamStatus.DISCONNECTED
            self.emit(EventType.DISCONNECTED)
        else:
            try:
                await self.generate_events_from_data(data)
            except Exception:
                self._logger.exception("Unable to process Afero IoT data. %s", data)

    async def __event_reader(self) -> None:
        """Poll the current states"""
        self._status = EventStreamStatus.CONNECTING
        while True:
            await self.perform_poll()
            await asyncio.sleep(self._polling_interval)

    async def process_event(self):
        try:
            event: AferoEvent = await self._event_queue.get()
            self.emit(event["type"], event)
        except Exception:
            self._logger.exception("Unhandled exception. Please open a bug report")

    async def __event_processor(self) -> None:
        """Process the Afero IoT devices"""
        while True:
            await self.process_event()



"""Test ThermostatController"""

import logging

import pytest

from aioafero.device import AferoState
from aioafero.v1.controllers.device import AferoBinarySensor
from aioafero.v1.controllers.thermostat import (
    ThermostatController,
    features,
    get_supported_modes,
)

from .. import utils

thermostat = utils.create_devices_from_data("thermostat.json")[0]
thermostat_id = "cc770a99-25da-4888-8a09-2a569da5be08"


@pytest.fixture
def mocked_controller(mocked_bridge, mocker):
    mocker.patch("time.time", return_value=12345)
    controller = ThermostatController(mocked_bridge)
    yield controller


@pytest.mark.asyncio
async def test_initialize(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    assert len(mocked_controller.items) == 1
    dev = mocked_controller.items[0]
    assert dev.id == thermostat_id
    assert dev.current_temperature == 18.3
    assert dev.fan_mode == features.ModeFeature(
        mode="auto", modes={"on", "auto", "intermittent"}
    )
    assert dev.fan_running is False
    assert dev.hvac_action == "off"
    assert dev.hvac_mode == features.HVACModeFeature(
        mode="heat",
        previous_mode="heat",
        modes={"off", "heat", "auto", "fan", "cool"},
        supported_modes={"off", "heat", "fan"},
    )
    assert dev.safety_max_temp == features.TargetTemperatureFeature(
        value=36, min=29.5, max=37, step=0.5, instance="safety-mode-max-temp"
    )
    assert dev.safety_min_temp == features.TargetTemperatureFeature(
        value=4, min=4, max=13, step=0.5, instance="safety-mode-min-temp"
    )
    assert dev.target_temperature_auto_cooling == features.TargetTemperatureFeature(
        value=26.5, step=0.5, min=10, max=37, instance="auto-cooling-target"
    )
    assert dev.target_temperature_auto_heating == features.TargetTemperatureFeature(
        value=18.5, step=0.5, min=4, max=32, instance="auto-heating-target"
    )
    assert dev.target_temperature_cooling == features.TargetTemperatureFeature(
        value=26.5, step=0.5, min=10, max=37, instance="cooling-target"
    )
    assert dev.target_temperature_heating == features.TargetTemperatureFeature(
        value=18, step=0.5, min=4, max=32, instance="heating-target"
    )
    assert dev.sensors == {}
    assert dev.binary_sensors == {
        "filter-replacement|None": AferoBinarySensor(
            id="filter-replacement|None",
            owner="cc770a99-25da-4888-8a09-2a569da5be08",
            _value="not-needed",
            _error="replacement-needed",
            instance=None,
        ),
        "max-temp-exceeded|None": AferoBinarySensor(
            id="max-temp-exceeded|None",
            owner="cc770a99-25da-4888-8a09-2a569da5be08",
            _value="normal",
            _error="alerting",
            instance=None,
        ),
        "min-temp-exceeded|None": AferoBinarySensor(
            id="min-temp-exceeded|None",
            owner="cc770a99-25da-4888-8a09-2a569da5be08",
            _value="normal",
            _error="alerting",
            instance=None,
        ),
    }


@pytest.mark.asyncio
async def test_update_elem(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    mocked_controller._items[thermostat_id].hvac_mode.mode = "heat"
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("thermostat.json")[0]
    new_states = [
        AferoState(
            **{
                "functionClass": "current-fan-state",
                "value": "on",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "fan-mode",
                "value": "on",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 22,
                "lastUpdateTime": 0,
                "functionInstance": "auto-heating-target",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 22.5,
                "lastUpdateTime": 0,
                "functionInstance": "auto-cooling-target",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 17,
                "lastUpdateTime": 0,
                "functionInstance": "heating-target",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 18,
                "lastUpdateTime": 0,
                "functionInstance": "cooling-target",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 19,
                "lastUpdateTime": 0,
                "functionInstance": "current-temp",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 35,
                "lastUpdateTime": 0,
                "functionInstance": "safety-mode-max-temp",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 32,
                "lastUpdateTime": 0,
                "functionInstance": "safety-mode-min-temp",
            }
        ),
        AferoState(
            **{
                "functionClass": "mode",
                "value": "cool",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "current-system-state",
                "value": "cooling",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "available",
                "value": False,
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.current_temperature == 19
    assert dev.fan_running is True
    assert dev.fan_mode.mode == "on"
    assert dev.hvac_action == "cooling"
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "heat"
    assert dev.safety_max_temp.value == 35
    assert dev.safety_min_temp.value == 32
    assert dev.target_temperature_auto_heating.value == 22
    assert dev.target_temperature_auto_cooling.value == 22.5
    assert dev.target_temperature_heating.value == 17
    assert dev.target_temperature_cooling.value == 18
    assert dev.available is False
    assert updates == {
        "temperature-safety-mode-max-temp",
        "temperature-heating-target",
        "temperature-auto-cooling-target",
        "temperature-cooling-target",
        "current-fan-state",
        "temperature-current-temp",
        "current-system-state",
        "available",
        "temperature-auto-heating-target",
        "mode",
        "temperature-safety-mode-min-temp",
        "fan-mode",
    }


@pytest.mark.asyncio
async def test_update_elem_no_prev_mode_change(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    mocked_controller._items[thermostat_id].hvac_mode.mode = "off"
    mocked_controller._items[thermostat_id].hvac_mode.previous_mode = "heat"
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("thermostat.json")[0]
    new_states = [
        AferoState(
            **{
                "functionClass": "mode",
                "value": "cool",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    dev = mocked_controller.items[0]
    assert dev.hvac_mode.mode == "cool"
    assert dev.hvac_mode.previous_mode == "heat"
    assert updates == {"mode"}


@pytest.mark.asyncio
async def test_update_elem_no_updates(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    mocked_controller._items[thermostat_id].hvac_mode.mode = "heat"
    assert len(mocked_controller.items) == 1
    dev_update = utils.create_devices_from_data("thermostat.json")[0]
    new_states = [
        AferoState(
            **{
                "functionClass": "current-fan-state",
                "value": "off",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "fan-mode",
                "value": "auto",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 18.5,
                "lastUpdateTime": 0,
                "functionInstance": "auto-heating-target",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 26.5,
                "lastUpdateTime": 0,
                "functionInstance": "auto-cooling-target",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 18,
                "lastUpdateTime": 0,
                "functionInstance": "heating-target",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 26.5,
                "lastUpdateTime": 0,
                "functionInstance": "cooling-target",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 18.3,
                "lastUpdateTime": 0,
                "functionInstance": "current-temp",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 36,
                "lastUpdateTime": 0,
                "functionInstance": "safety-mode-max-temp",
            }
        ),
        AferoState(
            **{
                "functionClass": "temperature",
                "value": 4,
                "lastUpdateTime": 0,
                "functionInstance": "safety-mode-min-temp",
            }
        ),
        AferoState(
            **{
                "functionClass": "mode",
                "value": "heat",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "current-system-state",
                "value": "off",
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
        AferoState(
            **{
                "functionClass": "available",
                "value": True,
                "lastUpdateTime": 0,
                "functionInstance": None,
            }
        ),
    ]
    for state in new_states:
        utils.modify_state(dev_update, state)
    updates = await mocked_controller.update_elem(dev_update)
    assert updates == set()


@pytest.mark.asyncio
async def test_set_state(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    mocked_controller._items[thermostat_id].hvac_mode.mode = "heat"
    mocked_controller._items[thermostat_id].hvac_mode.supported_modes.add("cool")
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(
        thermostat_id,
        hvac_mode="cool",
        safety_max_temp=35,
        safety_min_temp=8,
        target_temperature_auto_heating=22,
        target_temperature_auto_cooling=22.5,
        target_temperature_heating=17,
        target_temperature_cooling=18,
    )
    dev = mocked_controller.items[0]
    assert dev.fan_mode.mode == "auto"
    assert dev.fan_running is False
    assert dev.hvac_mode.mode == "cool"
    assert dev.safety_max_temp.value == 35
    assert dev.safety_min_temp.value == 8
    assert dev.target_temperature_auto_heating.value == 22
    assert dev.target_temperature_auto_cooling.value == 22.5
    assert dev.target_temperature_heating.value == 17
    assert dev.target_temperature_cooling.value == 18
    post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
    assert post["metadeviceId"] == thermostat_id
    expected_calls = [
        {
            "functionClass": "mode",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": "cool",
        },
        {
            "functionClass": "temperature",
            "functionInstance": "safety-mode-min-temp",
            "lastUpdateTime": 12345,
            "value": 8,
        },
        {
            "functionClass": "temperature",
            "functionInstance": "safety-mode-max-temp",
            "lastUpdateTime": 12345,
            "value": 35,
        },
        {
            "functionClass": "temperature",
            "functionInstance": "auto-cooling-target",
            "lastUpdateTime": 12345,
            "value": 22.5,
        },
        {
            "functionClass": "temperature",
            "functionInstance": "auto-heating-target",
            "lastUpdateTime": 12345,
            "value": 22,
        },
        {
            "functionClass": "temperature",
            "functionInstance": "heating-target",
            "lastUpdateTime": 12345,
            "value": 17,
        },
        {
            "functionClass": "temperature",
            "functionInstance": "cooling-target",
            "lastUpdateTime": 12345,
            "value": 18,
        },
    ]
    for call in expected_calls:
        assert call in post["values"]
    assert len(expected_calls) == len(post["values"])


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "current_mode, prev_mode, params, expected_calls, expected_messages",
    [
        # Testing target_temp / cooling
        (
            "cool",
            "cool",
            {"target_temperature": 25},
            [
                {
                    "functionClass": "temperature",
                    "functionInstance": "cooling-target",
                    "lastUpdateTime": 12345,
                    "value": 25,
                }
            ],
            [],
        ),
        # Testing target_temp / cooling
        (
            "heat",
            "heat",
            {"target_temperature": 24},
            [
                {
                    "functionClass": "temperature",
                    "functionInstance": "heating-target",
                    "lastUpdateTime": 12345,
                    "value": 24,
                }
            ],
            [],
        ),
        # Testing changing mode
        (
            "heat",
            "heat",
            {"target_temperature": 24, "hvac_mode": "cool"},
            [
                {
                    "functionClass": "temperature",
                    "functionInstance": "cooling-target",
                    "lastUpdateTime": 12345,
                    "value": 24,
                },
                {
                    "functionClass": "mode",
                    "functionInstance": None,
                    "lastUpdateTime": 12345,
                    "value": "cool",
                },
            ],
            [],
        ),
        # Testing target_temp / debug message
        (
            "off",
            "off",
            {"target_temperature": 25},
            [],
            ["Unable to set the target temperature due to the active mode: off"],
        ),
        # Invalid fan mode
        (
            "auto",
            "cool",
            {"fan_mode": "beans"},
            [],
            ["Unknown fan mode beans. Available modes: auto, intermittent, on"],
        ),
        # Invalid hvac mode
        (
            "auto",
            "cool",
            {"hvac_mode": "beans"},
            [],
            ["Unknown hvac mode beans. Available modes: auto, cool, fan, heat, off"],
        ),
    ],
)
async def test_set_state_hvac_generics(
    current_mode,
    prev_mode,
    params,
    expected_calls,
    expected_messages,
    mocked_controller,
    caplog,
):
    caplog.set_level(logging.DEBUG)
    await mocked_controller.initialize_elem(thermostat)
    mocked_controller._items[thermostat_id].hvac_mode.mode = current_mode
    mocked_controller._items[thermostat_id].hvac_mode.supported_modes.add("cool")
    mocked_controller._items[thermostat_id].hvac_mode.previous_mode = prev_mode
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(thermostat_id, **params)
    if expected_calls:
        post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
        assert post["metadeviceId"] == thermostat_id
        for call in expected_calls:
            assert call in post["values"]
        assert len(expected_calls) == len(post["values"])
    else:
        mocked_controller._bridge.request.assert_not_called()
    for message in expected_messages:
        assert message in caplog.text


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "fan_mode, hvac_mode, params, expected_calls, expected_messages",
    [
        (
            "off",
            "off",
            {"fan_mode": "auto"},
            [
                {
                    "functionClass": "fan-mode",
                    "functionInstance": None,
                    "lastUpdateTime": 12345,
                    "value": "auto",
                },
                {
                    "functionClass": "mode",
                    "functionInstance": None,
                    "lastUpdateTime": 12345,
                    "value": "fan",
                },
            ],
            [],
        ),
        (
            "on",
            "fan",
            {"fan_mode": "auto"},
            [
                {
                    "functionClass": "fan-mode",
                    "functionInstance": None,
                    "lastUpdateTime": 12345,
                    "value": "auto",
                },
                {
                    "functionClass": "mode",
                    "functionInstance": None,
                    "lastUpdateTime": 12345,
                    "value": "fan",
                },
            ],
            [],
        ),
        (
            "off",
            "off",
            {"fan_mode": "bad"},
            [],
            ["Unknown fan mode bad. Available modes: auto, intermittent, on"],
        ),
    ],
)
async def test_set_state_fan_generics(
    fan_mode,
    hvac_mode,
    params,
    expected_calls,
    expected_messages,
    mocked_controller,
    caplog,
):
    caplog.set_level(logging.DEBUG)
    await mocked_controller.initialize_elem(thermostat)
    mocked_controller._items[thermostat_id].hvac_mode.mode = hvac_mode
    mocked_controller._items[thermostat_id].fan_mode.mode = fan_mode
    assert len(mocked_controller.items) == 1
    await mocked_controller.set_state(thermostat_id, **params)
    if expected_calls:
        post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
        assert post["metadeviceId"] == thermostat_id
        for call in expected_calls:
            assert call in post["values"]
        assert len(expected_calls) == len(post["values"])
    else:
        mocked_controller._bridge.request.assert_not_called()
    for message in expected_messages:
        assert message in caplog.text


@pytest.mark.asyncio
async def test_set_fan_mode(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    assert len(mocked_controller.items) == 1
    mocked_controller._items[thermostat_id].fan_mode.mode = "off"
    await mocked_controller.set_fan_mode(thermostat_id, "on")
    post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
    assert post["metadeviceId"] == thermostat_id
    expected_calls = [
        {
            "functionClass": "fan-mode",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": "on",
        },
        {
            "functionClass": "mode",
            "functionInstance": None,
            "lastUpdateTime": 12345,
            "value": "fan",
        },
    ]
    assert len(post["values"]) == 2
    for call in expected_calls:
        assert call in post["values"]


@pytest.mark.asyncio
async def test_set_hvac_mode(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    assert len(mocked_controller.items) == 1
    mocked_controller._items[thermostat_id].hvac_mode.mode = "heat"
    mocked_controller._items[thermostat_id].hvac_mode.supported_modes.add("cool")
    await mocked_controller.set_hvac_mode(thermostat_id, "cool")
    post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
    assert post["metadeviceId"] == thermostat_id
    expected_call = {
        "functionClass": "mode",
        "functionInstance": None,
        "lastUpdateTime": 12345,
        "value": "cool",
    }
    assert len(post["values"]) == 1
    assert expected_call in post["values"]


@pytest.mark.asyncio
async def test_set_target_temperature(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    assert len(mocked_controller.items) == 1
    mocked_controller._items[thermostat_id].hvac_mode.mode = "heat"
    mocked_controller._items[thermostat_id].target_temperature_heating.value = 20
    await mocked_controller.set_target_temperature(thermostat_id, 21)
    post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
    assert post["metadeviceId"] == thermostat_id
    expected_call = {
        "functionClass": "temperature",
        "functionInstance": "heating-target",
        "lastUpdateTime": 12345,
        "value": 21,
    }
    assert len(post["values"]) == 1
    assert expected_call in post["values"]


@pytest.mark.asyncio
async def test_set_temperature_range(mocked_controller):
    await mocked_controller.initialize_elem(thermostat)
    assert len(mocked_controller.items) == 1
    mocked_controller._items[thermostat_id].hvac_mode.mode = "auto"
    mocked_controller._items[thermostat_id].target_temperature_auto_heating.value = 20
    mocked_controller._items[thermostat_id].target_temperature_auto_cooling.value = 21
    await mocked_controller.set_temperature_range(thermostat_id, 21, 22)
    post = mocked_controller._bridge.request.call_args_list[0][1]["json"]
    assert post["metadeviceId"] == thermostat_id
    expected_calls = [
        {
            "functionClass": "temperature",
            "functionInstance": "auto-cooling-target",
            "lastUpdateTime": 12345,
            "value": 22,
        },
        {
            "functionClass": "temperature",
            "functionInstance": "auto-heating-target",
            "lastUpdateTime": 12345,
            "value": 21,
        },
    ]
    for call in expected_calls:
        assert call in post["values"]
    assert len(post["values"]) == 2


standard = {"heat", "cool", "off", "fan", "auto"}


@pytest.mark.parametrize(
    "system_type, modes, expected",
    [
        ("cool-beans", {"heat", "cool", "beans"}, {"beans"}),
        ("1-compressor-heat-pump-1-aux-on-cool-boiler-aux", standard, standard),
        ("1-stage-cooling-conventional", standard, {"cool", "off", "fan"}),
        ("1-stage-heating-conventional-boiler", standard, {"heat", "off", "fan"}),
    ],
)
def test_get_supported_modes(system_type, modes, expected):
    assert get_supported_modes(system_type, modes) == expected



__all__ = ["AferoAuth"]

import asyncio
import base64
import contextlib
import datetime
import hashlib
import logging
import os
import re
from collections import namedtuple
from contextlib import contextmanager
from typing import Final, Optional
from urllib.parse import parse_qs, urlparse

import aiohttp
from aiohttp import ClientSession, ContentTypeError
from bs4 import BeautifulSoup
from securelogging import LogRedactorMessage, add_secret, remove_secret

from ..errors import InvalidAuth, InvalidResponse
from . import v1_const

logger = logging.getLogger(__name__)

TOKEN_TIMEOUT: Final[int] = 118
STATUS_CODE: Final[str] = "Status Code: %s"

auth_challenge = namedtuple("AuthChallenge", ["challenge", "verifier"])
token_data = namedtuple(
    "TokenData", ["token", "access_token", "refresh_token", "expiration"]
)
auth_sess_data = namedtuple("AuthSessionData", ["session_code", "execution", "tab_id"])


@contextmanager
def passthrough():
    yield


class AferoAuth:
    """Authentication against the Afero IoT API

    This class follows the Afero IoT authentication workflow and utilizes
    refresh tokens.
    """

    def __init__(
        self,
        username,
        password,
        hide_secrets: bool = True,
        refresh_token: Optional[str] = None,
        afero_client: Optional[str] = "hubspace",
    ):
        self.logger = logging.getLogger(f"{__package__}[{username}]")
        if hide_secrets:
            self.secret_logger = LogRedactorMessage
        else:
            self.secret_logger = passthrough
        self._hide_secrets: bool = hide_secrets
        self._async_lock: asyncio.Lock = asyncio.Lock()
        self._username: str = username
        self._password: str = password
        self._token_data: Optional[token_data] = None
        if refresh_token:
            add_secret(refresh_token)
            self._token_data = token_data(
                None, None, refresh_token, datetime.datetime.now().timestamp()
            )
        self._afero_client: str = afero_client
        self._token_headers: dict[str, str] = {
            "Content-Type": "application/x-www-form-urlencoded",
            "user-agent": v1_const.AFERO_CLIENTS[self._afero_client][
                "DEFAULT_USERAGENT"
            ],
            "host": v1_const.AFERO_CLIENTS[self._afero_client]["OPENID_HOST"],
        }

    @property
    async def is_expired(self) -> bool:
        """Determine if the token is expired"""
        if not self._token_data:
            return True
        return datetime.datetime.now().timestamp() >= self._token_data.expiration

    @property
    def refresh_token(self) -> Optional[str]:
        return self._token_data.refresh_token

    async def webapp_login(
        self, challenge: auth_challenge, client: ClientSession
    ) -> str:
        """Perform login to the webapp for a code

        Login to the webapp and generate a code used for generating tokens.

        :param challenge: Challenge data for connection and approving
        :param client: async client for making requests

        :return: Code used for generating a refresh token
        """
        code_params: dict[str, str] = {
            "response_type": "code",
            "client_id": v1_const.AFERO_CLIENTS[self._afero_client][
                "DEFAULT_CLIENT_ID"
            ],
            "redirect_uri": v1_const.AFERO_CLIENTS[self._afero_client][
                "DEFAULT_REDIRECT_URI"
            ],
            "code_challenge": challenge.challenge,
            "code_challenge_method": "S256",
            "scope": "openid offline_access",
        }
        self.logger.debug(
            "URL: %s\n\tparams: %s",
            v1_const.AFERO_CLIENTS[self._afero_client]["OPENID_URL"],
            code_params,
        )
        async with client.get(
            v1_const.AFERO_CLIENTS[self._afero_client]["OPENID_URL"],
            params=code_params,
            allow_redirects=False,
        ) as response:
            if response.status == 200:
                contents = await response.text()
                login_data = await extract_login_data(contents)
                self.logger.debug(
                    (
                        "WebApp Login:"
                        "\n\tSession Code: %s"
                        "\n\tExecution: %s"
                        "\n\tTab ID:%s"
                    ),
                    login_data.session_code,
                    login_data.execution,
                    login_data.tab_id,
                )
                return await self.generate_code(
                    login_data.session_code,
                    login_data.execution,
                    login_data.tab_id,
                    client,
                )
            elif response.status == 302:
                self.logger.debug("Hubspace returned an active session")
                return await AferoAuth.parse_code(response)
            else:
                raise InvalidResponse("Unable to query login page")

    @staticmethod
    async def generate_challenge_data() -> auth_challenge:
        code_verifier = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
        code_verifier = re.sub("[^a-zA-Z0-9]+", "", code_verifier)
        code_challenge = hashlib.sha256(code_verifier.encode("utf-8")).digest()
        code_challenge = base64.urlsafe_b64encode(code_challenge).decode("utf-8")
        code_challenge = code_challenge.replace("=", "")
        chal = auth_challenge(code_challenge, code_verifier)
        logger.debug("Challenge information: %s", chal)
        return chal

    async def generate_code(
        self, session_code: str, execution: str, tab_id: str, client: ClientSession
    ) -> str:
        """Finalize login to Afero IoT page

        :param session_code: Session code during form interaction
        :param execution: Session code during form interaction
        :param tab_id: Session code during form interaction
        :param client: async client for making request

        :return: code for generating tokens
        """
        self.logger.debug("Generating code")
        params = {
            "session_code": session_code,
            "execution": execution,
            "client_id": v1_const.AFERO_CLIENTS[self._afero_client][
                "DEFAULT_CLIENT_ID"
            ],
            "tab_id": tab_id,
        }
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "user-agent": v1_const.AFERO_CLIENTS[self._afero_client][
                "DEFAULT_USERAGENT"
            ],
        }
        auth_data = {
            "username": self._username,
            "password": self._password,
            "credentialId": "",
        }
        self.logger.debug(
            "URL: %s\n\tparams: %s\n\theaders: %s",
            v1_const.AFERO_CLIENTS[self._afero_client]["CODE_URL"],
            params,
            headers,
        )
        async with client.post(
            v1_const.AFERO_CLIENTS[self._afero_client]["CODE_URL"],
            params=params,
            data=auth_data,
            headers=headers,
            allow_redirects=False,
        ) as response:
            self.logger.debug(STATUS_CODE, response.status)
            if response.status != 302:
                raise InvalidAuth(
                    "Unable to authenticate with the supplied username / password"
                )
            return await AferoAuth.parse_code(response)

    @staticmethod
    async def parse_code(response: aiohttp.ClientResponse) -> str:
        """Parses the code for generating tokens"""
        try:
            parsed_url = urlparse(response.headers["location"])
            code = parse_qs(parsed_url.query)["code"][0]
            logger.debug("Location: %s", response.headers.get("location"))
            logger.debug("Code: %s", code)
        except KeyError:
            raise InvalidResponse(
                f"Unable to process the result from {response.url}: {response.status}"
            )
        return code

    async def generate_refresh_token(
        self,
        client: ClientSession,
        challenge: auth_challenge | None = None,
        code: str | None = None,
    ) -> token_data:
        """Generate a refresh token

        If a challenge is provided, it will send the correct data. If no challenge is required,
        it will use the existing token

        :param client: async client for making request
        :param code: Code used for generating refresh token
        :param challenge: Challenge data for connection and approving

        :return: Refresh token to generate a new token
        """
        self.logger.debug("Generating refresh token")
        if challenge:
            data = {
                "grant_type": "authorization_code",
                "code": code,
                "redirect_uri": v1_const.AFERO_CLIENTS[self._afero_client][
                    "DEFAULT_REDIRECT_URI"
                ],
                "code_verifier": challenge.verifier,
                "client_id": v1_const.AFERO_CLIENTS[self._afero_client][
                    "DEFAULT_CLIENT_ID"
                ],
            }
        else:
            data = {
                "grant_type": "refresh_token",
                "refresh_token": self._token_data.refresh_token,
                "scope": "openid email offline_access profile",
                "client_id": v1_const.AFERO_CLIENTS[self._afero_client][
                    "DEFAULT_CLIENT_ID"
                ],
            }
        with self.secret_logger():
            self.logger.debug(
                "URL: %s\n\tdata: %s\n\theaders: %s",
                v1_const.AFERO_CLIENTS[self._afero_client]["TOKEN_URL"],
                data,
                self._token_headers,
            )
        async with client.post(
            v1_const.AFERO_CLIENTS[self._afero_client]["TOKEN_URL"],
            headers=self._token_headers,
            data=data,
        ) as response:
            self.logger.debug(STATUS_CODE, response.status)
            with contextlib.suppress(ValueError, ContentTypeError):
                resp_json = await response.json()
            if response.status != 200:
                if resp_json and resp_json.get("error") == "invalid_grant":
                    raise InvalidAuth()
                else:
                    response.raise_for_status()
            try:
                refresh_token = resp_json["refresh_token"]
                access_token = resp_json["access_token"]
                id_token = resp_json["id_token"]
            except KeyError:
                raise InvalidResponse("Unable to extract refresh token")
            add_secret(refresh_token)
            add_secret(access_token)
            add_secret(id_token)
            with self.secret_logger():
                self.logger.debug("JSON response: %s", resp_json)
            return token_data(
                id_token,
                access_token,
                refresh_token,
                datetime.datetime.now().timestamp() + TOKEN_TIMEOUT,
            )

    async def perform_initial_login(self, client: ClientSession) -> token_data:
        """Login to generate a refresh token

        :param client: async client for making request

        :return: Refresh token for the auth
        """
        challenge = await AferoAuth.generate_challenge_data()
        code: str = await self.webapp_login(challenge, client)
        self.logger.debug("Successfully generated an auth code")
        refresh_token = await self.generate_refresh_token(
            client, code=code, challenge=challenge
        )
        self.logger.debug("Successfully generated a refresh token")
        return refresh_token

    async def token(self, client: ClientSession, retry: bool = True) -> str:
        invalidate_refresh_token = False
        async with self._async_lock:
            if not self._token_data:
                self.logger.debug(
                    "Refresh token not present. Generating a new refresh token"
                )
                self._token_data = await self.perform_initial_login(client)
            if await self.is_expired:
                self.logger.debug("Token has not been generated or is expired")
                try:
                    new_data = await self.generate_refresh_token(client)
                    remove_secret(self._token_data.token)
                    remove_secret(self._token_data.access_token)
                    remove_secret(self._token_data.refresh_token)
                    self._token_data = new_data
                except InvalidAuth:
                    self.logger.debug("Provided refresh token is no longer valid.")
                    if not retry:
                        raise
                    invalidate_refresh_token = True
                else:
                    self.logger.debug("Token has been successfully generated")
        if invalidate_refresh_token:
            return await self.token(client, retry=False)
        return self._token_data.token


async def extract_login_data(page: str) -> auth_sess_data:
    """Extract the required login data from the auth page

    :param page: the response from performing a GET against
    v1_const.AFERO_CLIENTS[self._afero_client]['OPENID_URL']
    """
    auth_page = BeautifulSoup(page, features="html.parser")
    login_form = auth_page.find("form", id="kc-form-login")
    if login_form is None:
        raise InvalidResponse("Unable to parse login page")
    try:
        login_url: str = login_form.attrs["action"]
    except KeyError:
        raise InvalidResponse("Unable to extract login url")
    parsed_url = urlparse(login_url)
    login_data = parse_qs(parsed_url.query)
    try:
        return auth_sess_data(
            login_data["session_code"][0],
            login_data["execution"][0],
            login_data["tab_id"][0],
        )
    except (KeyError, IndexError) as err:
        raise InvalidResponse("Unable to parse login url") from err



import asyncio
import copy
import re
import time
from asyncio.coroutines import iscoroutinefunction
from dataclasses import dataclass, fields
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Callable, Generic, Iterator

from ...device import AferoDevice, AferoResource, AferoState, get_afero_device
from ...errors import DeviceNotFound, ExceededMaximumRetries
from ...util import process_function
from .. import v1_const
from ..models.features import NumbersFeature, SelectFeature
from ..models.resource import ResourceTypes
from ..models.sensor import AferoBinarySensor, AferoSensor
from .event import AferoEvent, EventCallBackType, EventType

if TYPE_CHECKING:  # pragma: no cover
    from .. import AferoBridgeV1


EventSubscriptionType = tuple[
    EventCallBackType,
    "tuple[EventType] | None",
]

ID_FILTER_ALL = "*"

unit_extractor = re.compile(r"(\d*)(\D*)")


class BaseResourcesController(Generic[AferoResource]):
    """Base Controller for Afero IoT Cloud devices"""

    ITEM_TYPE_ID: ResourceTypes | None = None
    ITEM_TYPES: list[ResourceTypes] | None = None
    ITEM_CLS = None
    # functionClass map between controller -> Afero IoT Cloud
    ITEM_MAPPING: dict = {}
    # Sensors map functionClass -> Unit
    ITEM_SENSORS: dict[str, str] = {}
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {}
    # Elements that map to numbers. func class / func instance to unit
    ITEM_NUMBERS: dict[tuple[str, str | None], str] = {}
    # Elements that map to selects func class / func instance to name
    ITEM_SELECTS: dict[tuple[str, str | None], str] = {}

    def __init__(self, bridge: "AferoBridgeV1") -> None:
        """Initialize instance."""
        self._bridge = bridge
        self._items: dict[str, AferoResource] = {}
        self._logger = bridge.logger.getChild(self.ITEM_CLS.__name__)
        self._subscribers: dict[str, EventSubscriptionType] = {ID_FILTER_ALL: []}
        self._initialized: bool = False
        self._item_values = [x.value for x in self.ITEM_TYPES]

    def __getitem__(self, device_id: str) -> AferoResource:
        """Get item by device_id."""
        return self._items[device_id]

    def __iter__(self) -> Iterator[AferoResource]:
        """Iterate items."""
        return iter(self._items.values())

    def __contains__(self, device_id: str) -> bool:
        """Return bool if device_id is in items."""
        return device_id in self._items

    @property
    def items(self) -> list[AferoResource]:
        """Return all items for this resource."""
        return list(self._items.values())

    @property
    def initialized(self) -> bool:
        return self._initialized

    async def _handle_event(
        self, evt_type: EventType, evt_data: AferoEvent | None
    ) -> None:
        """Handle incoming event for this resource"""
        if evt_data is None:
            return
        item_id = evt_data.get("device_id", None)
        cur_item = await self._handle_event_type(evt_type, item_id, evt_data)
        if cur_item:
            await self.emit_to_subscribers(evt_type, item_id, cur_item)

    async def _handle_event_type(
        self, evt_type: EventType, item_id: str, evt_data: AferoEvent
    ) -> AferoResource | None:
        """Determines what to do with the incoming event

        :param evt_type: Type of event
        :param item_id: ID of the item
        :param evt_data: Event data

        :return: Item after being processed
        """
        if evt_type == EventType.RESOURCE_ADDED:
            self._logger.info(
                "Initializing %s as a %s", evt_data["device"].id, self.ITEM_CLS.__name__
            )
            cur_item = await self.initialize_elem(evt_data["device"])
            self._items[item_id] = cur_item
            self._bridge.add_device(evt_data["device"].id, self)
        elif evt_type == EventType.RESOURCE_DELETED:
            cur_item = self._items.pop(item_id, evt_data)
            self._bridge.remove_device(evt_data["device_id"])
        elif evt_type == EventType.RESOURCE_UPDATED:
            # existing item updated
            try:
                cur_item = self.get_device(item_id)
            except DeviceNotFound:
                return
            if not await self.update_elem(evt_data["device"]) and not evt_data.get(
                "force_forward", False
            ):
                return
        else:
            # Skip all other events
            return
        return cur_item

    async def emit_to_subscribers(
        self, evt_type: EventType, item_id: str, item: AferoResource
    ):
        """Emit updates to subscribers

        :param evt_type: Type of event
        :param item_id: ID of the item
        :param item: Item to emit to subscribers
        """
        subscribers = (
            self._subscribers.get(item_id, []) + self._subscribers[ID_FILTER_ALL]
        )
        for callback, event_filter in subscribers:
            if event_filter is not None and evt_type not in event_filter:
                continue
            # dispatch the full resource object to the callback
            if iscoroutinefunction(callback):
                asyncio.create_task(callback(evt_type, item))
            else:
                callback(evt_type, item)

    def get_filtered_devices(self, initial_data: list[dict]) -> list[AferoDevice]:
        valid_devices: list[AferoDevice] = []
        for element in initial_data:
            if element["typeId"] != self.ITEM_TYPE_ID.value:
                self._logger.debug(
                    "TypeID [%s] does not match %s",
                    element["typeId"],
                    self.ITEM_TYPE_ID.value,
                )
                continue
            device = get_afero_device(element)
            if device.device_class not in self._item_values:
                self._logger.debug(
                    "Device Class [%s] is not contained in %s",
                    device.device_class,
                    self._item_values,
                )
                continue
            valid_devices.append(device)
        return valid_devices

    async def _get_valid_devices(self, initial_data: list[dict]) -> list[AferoDevice]:
        return self.get_filtered_devices(initial_data)

    async def initialize(self, initial_data: list[dict]) -> None:
        """Initialize controller by fetching all items for this resource type from bridge."""
        if self._initialized:
            return
        valid_devices: list[AferoDevice] = await self._get_valid_devices(initial_data)
        for device in valid_devices:
            await self._handle_event(
                EventType.RESOURCE_ADDED,
                AferoEvent(
                    type=EventType.RESOURCE_ADDED,
                    device_id=device.id,
                    device=device,
                ),
            )
        # subscribe to item updates
        res_filter = tuple(x.value for x in self.ITEM_TYPES)
        if res_filter:
            self._bridge.events.subscribe(
                self._handle_event,
                resource_filter=res_filter,
            )
        else:
            # Subscribe to all events
            self._bridge.events.subscribe(
                self._handle_event,
            )
        self._initialized = True

    async def initialize_number(
        self, func_def: dict, state: AferoState
    ) -> tuple[tuple[str, str | None], NumbersFeature] | None:
        key = (state.functionClass, state.functionInstance)
        if key in self.ITEM_NUMBERS.keys():
            working_def = func_def["values"][0]
            return key, NumbersFeature(
                value=state.value,
                min=working_def["range"]["min"],
                max=working_def["range"]["max"],
                step=working_def["range"]["step"],
                name=working_def.get("name"),
                unit=self.ITEM_NUMBERS[key],
            )
        return None

    async def initialize_select(
        self, functions: list[dict], state: AferoState
    ) -> tuple[tuple[str, str | None], SelectFeature] | None:
        key = (state.functionClass, state.functionInstance)
        if key in self.ITEM_SELECTS.keys():
            return key, SelectFeature(
                selected=state.value,
                selects=set(
                    process_function(
                        functions, state.functionClass, state.functionInstance
                    )
                ),
                name=self.ITEM_SELECTS[key],
            )
        return None

    async def initialize_sensor(
        self, state: AferoState, child_id: str
    ) -> AferoSensor | AferoBinarySensor | None:
        """Initialize the sensor

        :param state: State to update
        :param child_id: device_id of the parent device
        """
        if state.functionClass in self.ITEM_SENSORS.keys():
            value, unit = await self.split_sensor_data(state)
            return AferoSensor(
                id=state.functionClass,
                owner=child_id,
                _value=value,
                unit=unit,
            )
        elif state.functionClass in self.ITEM_BINARY_SENSORS.keys():
            value, _ = await self.split_sensor_data(state)
            key = f"{state.functionClass}|{state.functionInstance}"
            return AferoBinarySensor(
                id=key,
                owner=child_id,
                instance=state.functionInstance,
                _value=value,
                _error=self.ITEM_BINARY_SENSORS[state.functionClass],
            )
        return None

    async def update_number(
        self, state: AferoState, cur_item: AferoResource
    ) -> str | None:
        """Update the number if its tracked and a change has been detected

        :param state: State to update
        :param cur_item: Current item to update
        :return: Identifier of the number that was updated or None
        """
        key = (state.functionClass, state.functionInstance)
        if key in self.ITEM_NUMBERS.keys():
            if cur_item.numbers[key].value != state.value:
                cur_item.numbers[key].value = state.value
                return f"number-{key}"
        return None

    async def update_select(
        self, state: AferoState, cur_item: AferoResource
    ) -> str | None:
        """Update the select if its tracked and a change has been detected

        :param state: State to update
        :param cur_item: Current item to update
        :return: Identifier of the select that was updated or None
        """
        key = (state.functionClass, state.functionInstance)
        if key in self.ITEM_SELECTS.keys():
            if cur_item.selects[key].selected != state.value:
                cur_item.selects[key].selected = state.value
                return f"select-{key}"
        return None

    async def update_sensor(
        self, state: AferoState, cur_item: AferoResource
    ) -> str | None:
        """Update the sensor if its tracked and a change has been detected

        :param state: State to update
        :param cur_item: Current item to update
        :return: Identifier of the sensor that was updated or None
        """
        if state.functionClass in self.ITEM_SENSORS.keys():
            value, _ = await self.split_sensor_data(state)
            if cur_item.sensors[state.functionClass]._value != value:
                cur_item.sensors[state.functionClass]._value = value
                return f"sensor-{state.functionClass}"
        elif state.functionClass in self.ITEM_BINARY_SENSORS.keys():
            value, _ = await self.split_sensor_data(state)
            key = f"{state.functionClass}|{state.functionInstance}"
            if cur_item.binary_sensors[key]._value != value:
                cur_item.binary_sensors[key]._value = value
                return f"binary-{key}"
        return None

    async def split_sensor_data(self, state: AferoState) -> tuple[Any, str | None]:
        if isinstance(state.value, str):
            match = unit_extractor.match(state.value)
            if match and match.group(1) and match.group(2):
                return int(match.group(1)), match.group(2)
        return state.value, self.ITEM_SENSORS.get(state.functionClass, None)

    async def initialize_elem(self, element: AferoDevice) -> None:  # pragma: no cover
        raise NotImplementedError("Class should implement initialize_elem")

    async def update_elem(self, element: AferoDevice) -> None:  # pragma: no cover
        raise NotImplementedError("Class should implement update_elem")

    def subscribe(
        self,
        callback: EventCallBackType,
        id_filter: str | tuple[str] | None = None,
        event_filter: EventType | tuple[EventType] | None = None,
    ) -> Callable:
        """
        Subscribe to status changes for this resource type.

        Parameters:
            - `callback` - callback function to call when an event emits.
            - `id_filter` - Optionally provide resource ID(s) to filter events for.
            - `event_filter` - Optionally provide EventType(s) as filter.

        Returns:
            function to unsubscribe.
        """
        if not isinstance(event_filter, None | list | tuple):
            event_filter = (event_filter,)

        if id_filter is None:
            id_filter = (ID_FILTER_ALL,)
        elif not isinstance(id_filter, list | tuple):
            id_filter = (id_filter,)

        subscription = (callback, event_filter)

        for id_key in id_filter:
            if id_key not in self._subscribers:
                self._subscribers[id_key] = []
            self._subscribers[id_key].append(subscription)

        # unsubscribe logic
        def unsubscribe():
            for id_key in id_filter:
                if id_key not in self._subscribers:
                    continue
                self._subscribers[id_key].remove(subscription)

        return unsubscribe

    async def _process_state_update(
        self, cur_item: AferoResource, device_id: str, states: list[dict]
    ) -> None:
        dev_states = []
        for state in states:
            dev_states.append(
                AferoState(
                    functionClass=state["functionClass"],
                    value=state["value"],
                    functionInstance=state.get("functionInstance"),
                    lastUpdateTime=int(datetime.now(timezone.utc).timestamp() * 1000),
                )
            )
        dummy_update = AferoDevice(
            id=device_id,
            device_id=cur_item.device_information.parent_id,
            model=cur_item.device_information.model,
            device_class=cur_item.device_information.device_class,
            default_image=cur_item.device_information.default_image,
            default_name=cur_item.device_information.default_name,
            friendly_name=cur_item.device_information.name,
            states=dev_states,
        )
        # Update now, but also trigger all chained updates
        await self.update_elem(dummy_update)
        self._bridge.events.add_job(
            AferoEvent(
                type=EventType.RESOURCE_UPDATED,
                device_id=device_id,
                device=dummy_update,
                force_forward=True,
            )
        )

    async def update_afero_api(self, device_id: str, states: list[dict]) -> bool:
        """Update Afero IoT API with the new states

        :param device_id: Afero IoT Device ID
        :param states: States to manually set

        :return: True if successful, False otherwise.
        """
        url = v1_const.AFERO_CLIENTS[self._bridge.afero_client]["DEVICE_STATE"].format(
            self._bridge.account_id, str(device_id)
        )
        headers = {
            "host": v1_const.AFERO_CLIENTS[self._bridge.afero_client]["DATA_HOST"],
            "content-type": "application/json; charset=utf-8",
        }
        payload = {"metadeviceId": str(device_id), "values": states}
        try:
            res = await self._bridge.request("put", url, json=payload, headers=headers)
        except ExceededMaximumRetries:
            self._logger.warning("Maximum retries exceeded for %s", device_id)
            return False
        else:
            # Bad states provided
            if res.status == 400:
                self._logger.warning(
                    "Invalid update provided for %s using %s", device_id, states
                )
                return False
        return True

    async def update(
        self,
        device_id: str,
        obj_in: Generic[AferoResource] = None,
        states: list[dict] | None = None,
    ) -> None:
        """Update Afero IoT with the new data

        :param device_id: Afero IoT Device ID
        :param obj_in: Afero IoT Resource elements to change
        :param states: States to manually set
        """
        try:
            cur_item = self.get_device(device_id)
        except DeviceNotFound:
            self._logger.info(
                "Unable to update device %s as it does not exist", device_id
            )
            return
        # Make a clone to restore if the update fails
        fallback = copy.deepcopy(cur_item)
        if obj_in:
            device_states = dataclass_to_afero(cur_item, obj_in, self.ITEM_MAPPING)
            if not device_states:
                self._logger.debug("No states to send. Skipping")
                return
            # Update the state of the item to match the new states
            update_dataclass(cur_item, obj_in)
        else:  # Manually setting states
            device_states = states
            await self._process_state_update(cur_item, device_id, states)
        # @TODO - Implement bluetooth logic for update
        if not await self.update_afero_api(device_id, device_states):
            self._items[device_id] = fallback

    def get_device(self, device_id) -> AferoResource:
        try:
            return self[device_id]
        except KeyError:
            raise DeviceNotFound(device_id)


def update_dataclass(elem: AferoResource, update_vals: dataclass):
    """Updates the element with the latest changes"""
    for f in fields(update_vals):
        cur_val = getattr(update_vals, f.name, None)
        elem_val = getattr(elem, f.name)
        if cur_val is None:
            continue
        # There is probably a better way to approach this
        if not str(f.type).startswith("dict"):
            # Special processing for dicts
            if isinstance(elem_val, dict):
                cur_val = {getattr(cur_val, "func_instance", None): cur_val}
                getattr(elem, f.name).update(cur_val)
            else:
                setattr(elem, f.name, cur_val)
        else:
            elem_val.update(cur_val)


def dataclass_to_afero(
    elem: AferoResource, cls: dataclass, mapping: dict
) -> list[dict]:
    """Convert the current state to be consumed by Afero IoT"""
    states = []
    for f in fields(cls):
        current_feature = getattr(cls, f.name, None)
        if current_feature is None:
            continue
        api_key = mapping.get(f.name, f.name)
        # There is probably a better way to approach this
        field_is_dict = str(f.type).startswith("dict")
        is_tuple_key = False
        if field_is_dict and current_feature and current_feature.keys():
            is_tuple_key = isinstance(list(current_feature.keys())[0], tuple)
        # Tuple keys signify (func_class / func_instance).
        if field_is_dict and is_tuple_key:
            states.extend(get_afero_states_from_mapped(elem, f.name, current_feature))
        elif field_is_dict and not current_feature:
            continue
        else:
            # We need to determine funcClass / funcInstance when we dump our data
            if current_feature == getattr(elem, f.name, None):
                continue
            current_feature_value = current_feature
            if hasattr(current_feature, "api_value"):
                current_feature_value = current_feature.api_value
            if not isinstance(current_feature_value, list):
                func_instance = get_afero_instance_for_state(
                    elem, current_feature, api_key
                )
                states.append(
                    get_afero_state_from_feature(
                        api_key, func_instance, current_feature_value
                    )
                )
            else:
                states.extend(get_afero_states_from_list(current_feature_value))
    return states


def get_afero_states_from_mapped(
    element: AferoResource, field_name: str, update_vals: dict
) -> list[dict]:
    states = []
    current_elems = getattr(element, field_name, None)
    for key, val in update_vals.items():
        if val == current_elems.get(key, None):
            continue
        states.append(
            {
                "functionClass": key[0],
                "functionInstance": key[1],
                "lastUpdateTime": int(time.time()),
                "value": val.api_value,
            }
        )
    return states


def get_afero_instance_for_state(
    elem: AferoResource, feature, mapped_afero_key: str | None
) -> str | None:
    """Determine the function instance based on the field data or device"""
    if hasattr(feature, "func_instance") and getattr(feature, "func_instance", None):
        instance = getattr(feature, "func_instance", None)
    elif (
        mapped_afero_key
        and hasattr(elem, "get_instance")
        and elem.get_instance(mapped_afero_key)
    ):
        instance = elem.get_instance(mapped_afero_key)
    else:
        instance = None
    return instance


def get_afero_state_from_feature(
    func_class: str, func_instance: str | None, current_val: Any
) -> dict:
    """Generate a single state from the current data"""
    new_state = {
        "functionClass": func_class,
        "functionInstance": func_instance,
        "lastUpdateTime": int(time.time()),
        "value": None,
    }
    if isinstance(current_val, dict):
        new_state.update(current_val)
    else:
        new_state["value"] = current_val
    return new_state


def get_afero_states_from_list(states: list[dict]) -> list[dict]:
    """Add timestamp to the states

    Assume the state already has functionClass, functionState, and value
    """
    for state in states:
        state["lastUpdateTime"] = int(time.time())
    return states



import pytest

from aioafero.v1.models import features
from aioafero.v1.models.thermostat import Thermostat


@pytest.fixture
def populated_entity():
    return Thermostat(
        [
            {
                "functionClass": "preset",
                "functionInstance": "preset-1",
                "value": "on",
                "lastUpdateTime": 0,
            }
        ],
        id="entity-1",
        available=True,
        current_temperature=12,
        fan_mode=features.ModeFeature(mode="off", modes={"on", "off"}),
        hvac_action="off",
        hvac_mode=features.HVACModeFeature(
            mode="heat",
            previous_mode="heat",
            modes={"off", "heat", "auto", "fan", "cool"},
            supported_modes={"off", "heat", "auto", "fan", "cool"},
        ),
        safety_max_temp=features.TargetTemperatureFeature(
            value=36, step=0.5, min=29.5, max=37, instance="safety-mode-max-temp"
        ),
        safety_min_temp=features.TargetTemperatureFeature(
            value=4, step=0.5, min=4, max=13, instance="safety-mode-min-temp"
        ),
        target_temperature_auto_heating=features.TargetTemperatureFeature(
            value=18, step=0.5, min=4, max=32, instance="auto-heating-target"
        ),
        target_temperature_auto_cooling=features.TargetTemperatureFeature(
            value=26.5, step=0.5, min=4, max=32, instance="auto-cooling-target"
        ),
        target_temperature_heating=features.TargetTemperatureFeature(
            value=19, step=0.5, min=4, max=32, instance="heating-target"
        ),
        target_temperature_cooling=features.TargetTemperatureFeature(
            value=26, step=0.5, min=10, max=37, instance="cooling-target"
        ),
        instances="i dont execute",
    )


@pytest.fixture
def empty_entity():
    return Thermostat(
        [],
        id="entity-1",
        available=True,
        current_temperature=None,
        fan_mode=None,
        hvac_action=None,
        hvac_mode=None,
        safety_max_temp=None,
        safety_min_temp=None,
        target_temperature_auto_heating=None,
        target_temperature_auto_cooling=None,
        target_temperature_heating=None,
        target_temperature_cooling=None,
        instances="i dont execute",
    )


@pytest.mark.parametrize(
    "mode,prev_mode,expected",
    [
        ("cool", "cool", "cool"),
        ("heat", "heat", "heat"),
        ("off", "cool", "cool"),
        ("off", "heat", "heat"),
        ("auto", "cool", "cool"),
        ("auto", "heat", "heat"),
    ],
)
def test__get_mode_to_check(mode, prev_mode, expected, populated_entity):
    populated_entity.hvac_mode.mode = mode
    populated_entity.hvac_mode.previous_mode = prev_mode
    assert populated_entity.get_mode_to_check() == expected


def test__get_mode_to_not_populated(populated_entity):
    populated_entity.hvac_mode = None
    assert populated_entity.get_mode_to_check() is None


def test_target_temperature_step_not_populated(populated_entity):
    populated_entity.hvac_mode = None
    assert populated_entity.target_temperature_step == 0.5


def test_init(populated_entity):
    assert populated_entity.id == "entity-1"
    assert populated_entity.available is True
    assert populated_entity.instances == {"preset": "preset-1"}
    assert populated_entity.current_temperature == 12
    assert populated_entity.fan_mode.mode == "off"
    assert populated_entity.hvac_action == "off"
    assert populated_entity.hvac_mode.mode == "heat"
    assert populated_entity.safety_max_temp.value == 36
    assert populated_entity.safety_max_temp.instance == "safety-mode-max-temp"
    assert populated_entity.safety_min_temp.value == 4
    assert populated_entity.safety_min_temp.instance == "safety-mode-min-temp"
    assert populated_entity.target_temperature_auto_heating.value == 18
    assert populated_entity.target_temperature_auto_heating.step == 0.5
    assert populated_entity.target_temperature_auto_heating.min == 4
    assert populated_entity.target_temperature_auto_heating.max == 32
    assert (
        populated_entity.target_temperature_auto_heating.instance
        == "auto-heating-target"
    )
    # Property checks
    assert populated_entity.target_temperature == 19
    assert populated_entity.target_temperature_max == 32
    assert populated_entity.target_temperature_min == 4
    assert populated_entity.target_temperature_step == 0.5
    populated_entity.hvac_mode.mode = "auto"
    assert populated_entity.target_temperature is None
    assert populated_entity.target_temperature_max == 32
    assert populated_entity.target_temperature_min == 4
    assert populated_entity.target_temperature_step == 0.5
    assert populated_entity.target_temperature_range == (
        18,
        26.5,
    )
    assert populated_entity.supports_fan_mode
    assert populated_entity.supports_temperature_range
    populated_entity.hvac_mode.mode = "cool"
    assert populated_entity.target_temperature == 26
    assert populated_entity.target_temperature_max == 37
    assert populated_entity.target_temperature_min == 10
    assert populated_entity.target_temperature_step == 0.5
    populated_entity.hvac_mode.mode = "off"
    populated_entity.hvac_mode.previous_mode = "cool"
    assert populated_entity.target_temperature == 26
    assert populated_entity.target_temperature_max == 37
    assert populated_entity.target_temperature_min == 10
    assert populated_entity.target_temperature_step == 0.5
    populated_entity.hvac_mode.previous_mode = "heat"
    assert populated_entity.target_temperature == 19
    populated_entity.hvac_mode.previous_mode = "i-dont-exist"
    assert populated_entity.target_temperature is None


def test_init_empty(empty_entity):
    assert not empty_entity.current_temperature
    assert not empty_entity.fan_mode
    assert not empty_entity.hvac_action
    assert not empty_entity.hvac_mode
    assert not empty_entity.safety_max_temp
    assert not empty_entity.safety_min_temp
    assert not empty_entity.target_temperature_auto_heating

    # Property Checks
    assert not empty_entity.supports_fan_mode
    assert not empty_entity.supports_temperature_range


def test_get_instance(populated_entity):
    assert populated_entity.get_instance("preset") == "preset-1"



import logging

import pytest

from aioafero import EventType, InvalidAuth
from aioafero.errors import DeviceNotFound
from aioafero.v1 import AferoBridgeV1, add_secret
from aioafero.v1.controllers.device import DeviceController
from aioafero.v1.controllers.event import EventStream
from aioafero.v1.controllers.fan import FanController
from aioafero.v1.controllers.light import LightController
from aioafero.v1.controllers.lock import LockController
from aioafero.v1.controllers.switch import SwitchController
from aioafero.v1.controllers.valve import ValveController

from . import utils

zandra_light = utils.create_devices_from_data("fan-ZandraFan.json")[1]


@pytest.mark.skip(reason="Not yet implemented")
def test_context_manager(mocked_bridge):
    pass


def test_devices(mocked_bridge):
    assert isinstance(mocked_bridge.devices, DeviceController)


def test_events(mocked_bridge):
    assert isinstance(mocked_bridge.events, EventStream)


def test_fans(mocked_bridge):
    assert isinstance(mocked_bridge.fans, FanController)


def test_lights(mocked_bridge):
    assert isinstance(mocked_bridge.lights, LightController)


def test_locks(mocked_bridge):
    assert isinstance(mocked_bridge.locks, LockController)


def test_switches(mocked_bridge):
    assert isinstance(mocked_bridge.switches, SwitchController)


def test_valves(mocked_bridge):
    assert isinstance(mocked_bridge.valves, ValveController)


def test_controllers(mocked_bridge):
    mocked_bridge.devices._initialized = False
    mocked_bridge.exhaust_fans._initialized = False
    mocked_bridge.fans._initialized = False
    mocked_bridge.lights._initialized = False
    mocked_bridge.locks._initialized = True
    mocked_bridge.switches._initialized = False
    mocked_bridge.thermostats._initialized = False
    mocked_bridge.valves._initialized = True
    assert mocked_bridge.controllers == [mocked_bridge.locks, mocked_bridge.valves]
    mocked_bridge.switches._initialized = True
    assert mocked_bridge.controllers == [
        mocked_bridge.locks,
        mocked_bridge.switches,
        mocked_bridge.valves,
    ]


def test_tracked_devices(mocked_bridge):
    assert mocked_bridge.tracked_devices == set()
    mocked_bridge.add_device(zandra_light.id, mocked_bridge.lights)
    assert mocked_bridge.tracked_devices == {zandra_light.id}


def test_add_device(mocked_bridge):
    assert mocked_bridge.tracked_devices == set()
    mocked_bridge.add_device(zandra_light.id, mocked_bridge.lights)
    assert mocked_bridge.tracked_devices == {zandra_light.id}
    assert mocked_bridge._known_devs == {zandra_light.id: mocked_bridge.lights}


def test_remove_device(mocked_bridge):
    assert mocked_bridge.tracked_devices == set()
    mocked_bridge.add_device(zandra_light.id, mocked_bridge.lights)
    assert mocked_bridge.tracked_devices == {zandra_light.id}
    assert mocked_bridge._known_devs == {zandra_light.id: mocked_bridge.lights}
    mocked_bridge.remove_device(zandra_light.id)
    assert mocked_bridge.tracked_devices == set()
    assert mocked_bridge._known_devs == {}


def test_set_polling_interval(mocked_bridge):
    assert mocked_bridge.events._polling_interval == 30
    mocked_bridge.set_polling_interval(10)
    assert mocked_bridge.events._polling_interval == 10


@pytest.mark.skip(reason="Not yet implemented")
@pytest.mark.asyncio
async def test_close():
    pass


@pytest.mark.skip(reason="Not yet implemented")
def test_subscribe():
    pass


@pytest.mark.skip(reason="Not yet implemented")
@pytest.mark.asyncio
async def test_get_account_id():
    pass


@pytest.mark.skip(reason="Not yet implemented")
@pytest.mark.asyncio
async def test_initialize():
    pass


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "expected_val,error",
    [
        # good data
        ([], False),
        # bad data
        ("i dont know", True),
    ],
)
async def test_fetch_data(expected_val, error, mocked_bridge_req, mocker):
    expected = mocker.Mock()
    mocker.patch.object(
        expected, "json", side_effect=mocker.AsyncMock(return_value=expected_val)
    )
    mocker.patch.object(mocked_bridge_req, "request", return_value=expected)
    if not error:
        assert await mocked_bridge_req.fetch_data() == expected_val
    else:
        with pytest.raises(ValueError):
            await mocked_bridge_req.fetch_data()


@pytest.mark.asyncio
async def test_send_service_request_dev_not_found(mocked_bridge):
    with pytest.raises(DeviceNotFound):
        await mocked_bridge.send_service_request("no", [{}])


@pytest.mark.asyncio
async def test_send_service_request(mocked_bridge):
    controller = mocked_bridge.lights
    await controller.initialize_elem(zandra_light)
    mocked_bridge.add_device(zandra_light.id, controller)
    assert controller[zandra_light.id].on.on is True
    await mocked_bridge.send_service_request(
        zandra_light.id,
        [{"functionClass": "power", "functionInstance": "light-power", "value": "off"}],
    )
    assert controller[zandra_light.id].on.on is False


@pytest.mark.asyncio
async def test_create_request_err(mocked_bridge, mocker):
    mocker.patch.object(mocked_bridge._auth, "token", side_effect=InvalidAuth)
    emit = mocker.patch.object(mocked_bridge.events, "emit")
    with pytest.raises(InvalidAuth):
        async with mocked_bridge.create_request("get", "https://not-called.io"):
            pass

    emit.assert_called_once_with(EventType.INVALID_AUTH)


@pytest.mark.parametrize(
    "hide_secrets",
    [True, False],
)
def test_AferoBridgeV1_hide_secrets(hide_secrets, caplog):
    caplog.set_level(logging.DEBUG)
    bridge = AferoBridgeV1("username", "password", hide_secrets=hide_secrets)
    secret = "this-is-super-secret-beans"
    add_secret(secret)
    with bridge.secret_logger():
        bridge.logger.debug(secret)
    if hide_secrets:
        assert "th***ns" in caplog.text
    else:
        assert secret in caplog.text


class DummyResponse:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    async def read(self):
        return "cool-beans"


def double_429(*args, **kwargs):
    yield DummyResponse(status_code=429)
    yield DummyResponse(status_code=429)
    yield DummyResponse(status_code=200)


# @pytest.mark.asyncio
# @pytest.mark.parametrize(
#     "max_retries, times_to_sleep, response_gen, exp_error", [
#         # Out of retries
#         (
#             0, 0, None, ExceededMaximumRetries
#         ),
#         # Double retry
#         (
#             None, 2, double_429, None
#         )
#     ]
# )
# async def test_request(max_retries, times_to_sleep, response_gen, exp_error, mocker):
#     bridge = AferoBridgeV1("username", "password")
#     mock_sleep = mocker.patch("asyncio.sleep")
#     if response_gen:
#         mocker.patch.object(bridge, "create_request", side_effect=response_gen())
#     if max_retries is not None:
#         mocker.patch.object(v1_const, "MAX_RETRIES", max_retries)
#     if exp_error:
#         with pytest.raises(exp_error):
#             await bridge.request("fff", "fff")
#     if times_to_sleep:
#         assert mock_sleep.call_count == times_to_sleep
#



"""Controller holding and managing Afero IoT resources of type `switch`."""

from ... import errors
from ...device import AferoDevice
from ..models import features
from ..models.resource import DeviceInformation, ResourceTypes
from ..models.switch import Switch, SwitchPut
from .base import AferoBinarySensor, AferoSensor, BaseResourcesController


class SwitchController(BaseResourcesController[Switch]):
    """Controller holding and managing Afero IoT resources of type `switch`.

    A switch can have one or more toggleable elements. They are controlled
    by their functionInstance.
    """

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [
        ResourceTypes.SWITCH,
        ResourceTypes.POWER_OUTLET,
        ResourceTypes.LANDSCAPE_TRANSFORMER,
        # Exhaust-Fans have On / Off toggles
        ResourceTypes.EXHAUST_FAN,
    ]
    ITEM_CLS = Switch
    ITEM_MAPPING = {}
    # Sensors map functionClass -> Unit
    ITEM_SENSORS: dict[str, str] = {"watts": "W", "output-voltage-switch": "V"}
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {}

    async def turn_on(self, device_id: str, instance: str | None = None) -> None:
        """Turn on the switch."""
        await self.set_state(device_id, on=True, instance=instance)

    async def turn_off(self, device_id: str, instance: str | None = None) -> None:
        """Turn off the switch."""
        await self.set_state(device_id, on=False, instance=instance)

    async def initialize_elem(self, afero_device: AferoDevice) -> Switch:
        """Initialize the element"""
        available: bool = False
        on: dict[str, features.OnFeature] = {}
        sensors: dict[str, AferoSensor] = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        toggle_states = ["power", "toggle"]
        if afero_device.device_class == ResourceTypes.EXHAUST_FAN.value:
            toggle_states.remove("power")
        for state in afero_device.states:
            if state.functionClass in toggle_states:
                on[state.functionInstance] = features.OnFeature(
                    on=state.value == "on",
                    func_class=state.functionClass,
                    func_instance=state.functionInstance,
                )
            elif state.functionClass == "available":
                available = state.value
            elif sensor := await self.initialize_sensor(state, afero_device.device_id):
                if isinstance(sensor, AferoBinarySensor):
                    binary_sensors[sensor.id] = sensor
                else:
                    sensors[sensor.id] = sensor

        self._items[afero_device.id] = Switch(
            afero_device.functions,
            id=afero_device.id,
            available=available,
            sensors=sensors,
            binary_sensors=binary_sensors,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
            ),
            on=on,
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        cur_item = self.get_device(afero_device.id)
        updated_keys = set()
        toggle_states = ["power", "toggle"]
        if afero_device.device_class == ResourceTypes.EXHAUST_FAN.value:
            toggle_states.remove("power")
        for state in afero_device.states:
            if state.functionClass in toggle_states:
                new_val = state.value == "on"
                if cur_item.on[state.functionInstance].on != new_val:
                    updated_keys.add("on")
                cur_item.on[state.functionInstance].on = state.value == "on"
            elif state.functionClass == "available":
                if cur_item.available != state.value:
                    updated_keys.add("available")
                cur_item.available = state.value
            elif update_key := await self.update_sensor(state, cur_item):
                updated_keys.add(update_key)

        return updated_keys

    async def set_state(
        self,
        device_id: str,
        on: bool | None = None,
        instance: str | None = None,
    ) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = SwitchPut()
        try:
            cur_item = self.get_device(device_id)
        except errors.DeviceNotFound:
            self._logger.info("Unable to find device %s", device_id)
            return
        if on is not None:
            try:
                update_obj.on = features.OnFeature(
                    on=on,
                    func_class=cur_item.on[instance].func_class,
                    func_instance=instance,
                )
            except KeyError:
                self._logger.info("Unable to find instance %s", instance)
        await self.update(device_id, obj_in=update_obj)



"""Controller holding and managing Afero IoT resources of type `exhaust-fan`."""

from ...device import AferoDevice, get_function_from_device
from ...errors import DeviceNotFound
from ..models import features
from ..models.exhaust_fan import ExhaustFan, ExhaustFanPut
from ..models.features import NumbersFeature, SelectFeature
from ..models.resource import DeviceInformation, ResourceTypes
from .base import AferoBinarySensor, AferoSensor, BaseResourcesController


class ExhaustFanController(BaseResourcesController[ExhaustFan]):
    """Controller holding and managing Afero IoT resources of type `exhaust-fan`.

    An exhaust fan tracks sensors, numbers, and selects. Toggles are controlled by
    SwitchController, fan is controlled by FanController, and light is controlled
    by LightController.
    """

    ITEM_TYPE_ID = ResourceTypes.DEVICE
    ITEM_TYPES = [ResourceTypes.EXHAUST_FAN]
    ITEM_CLS = ExhaustFan
    ITEM_MAPPING = {}
    # Sensors map functionClass -> Unit
    ITEM_SENSORS: dict[str, str] = {}
    # Binary sensors map key -> alerting value
    ITEM_BINARY_SENSORS: dict[str, str] = {
        "motion-detection": "motion-detected",
        "humidity-threshold-met": "above-threshold",
    }
    # Elements that map to numbers. func class / func instance to unit
    ITEM_NUMBERS: dict[tuple[str, str | None], str] = {
        ("auto-off-timer", "auto-off"): "seconds",
    }
    ITEM_SELECTS = {
        ("motion-action", "exhaust-fan"): "Motion Action",
        ("sensitivity", "humidity-sensitivity"): "Humidity Sensitivity",
    }

    async def initialize_elem(self, afero_device: AferoDevice) -> ExhaustFan:
        """Initialize the element"""
        self._logger.info("Initializing %s", afero_device.id)
        available: bool = False
        numbers: dict[tuple[str, str], features.NumbersFeature] = {}
        selects: dict[tuple[str, str], features.SelectFeature] = {}
        sensors: dict[str, AferoSensor] = {}
        binary_sensors: dict[str, AferoBinarySensor] = {}
        for state in afero_device.states:
            func_def = get_function_from_device(
                afero_device.functions, state.functionClass, state.functionInstance
            )
            if number := await self.initialize_number(func_def, state):
                numbers[number[0]] = number[1]
            elif select := await self.initialize_select(afero_device.functions, state):
                selects[select[0]] = select[1]
            elif state.functionClass == "available":
                available = state.value
            elif sensor := await self.initialize_sensor(state, afero_device.device_id):
                binary_sensors[sensor.id] = sensor

        self._items[afero_device.id] = ExhaustFan(
            afero_device.functions,
            id=afero_device.id,
            available=available,
            sensors=sensors,
            binary_sensors=binary_sensors,
            device_information=DeviceInformation(
                device_class=afero_device.device_class,
                default_image=afero_device.default_image,
                default_name=afero_device.default_name,
                manufacturer=afero_device.manufacturerName,
                model=afero_device.model,
                name=afero_device.friendly_name,
                parent_id=afero_device.device_id,
            ),
            numbers=numbers,
            selects=selects,
        )
        return self._items[afero_device.id]

    async def update_elem(self, afero_device: AferoDevice) -> set:
        cur_item = self.get_device(afero_device.id)
        updated_keys = set()
        for state in afero_device.states:
            if state.functionClass == "available":
                if cur_item.available != state.value:
                    updated_keys.add("available")
                cur_item.available = state.value
            elif update_key := await self.update_sensor(state, cur_item):
                updated_keys.add(update_key)
            elif update_key := await self.update_number(state, cur_item):
                updated_keys.add(update_key)
            elif update_key := await self.update_select(state, cur_item):
                updated_keys.add(update_key)
        return updated_keys

    async def set_state(
        self,
        device_id: str,
        numbers: dict[tuple[str, str], float] | None = None,
        selects: dict[tuple[str, str], str] | None = None,
    ) -> None:
        """Set supported feature(s) to fan resource."""
        update_obj = ExhaustFanPut()
        try:
            cur_item = self.get_device(device_id)
        except DeviceNotFound:
            self._logger.info("Unable to find device %s", device_id)
            return
        if numbers:
            for key, val in numbers.items():
                if key not in cur_item.numbers:
                    continue
                update_obj.numbers[key] = NumbersFeature(
                    value=val,
                    min=cur_item.numbers[key].min,
                    max=cur_item.numbers[key].max,
                    step=cur_item.numbers[key].step,
                    name=cur_item.numbers[key].name,
                    unit=cur_item.numbers[key].unit,
                )
        if selects:
            for key, val in selects.items():
                if key not in cur_item.selects:
                    continue
                update_obj.selects[key] = SelectFeature(
                    selected=val,
                    selects=cur_item.selects[key].selects,
                    name=cur_item.selects[key].name,
                )

        await self.update(device_id, obj_in=update_obj)



from enum import Enum


class EventType(Enum):
    """Enum with possible Events."""

    RESOURCE_ADDED = "add"
    RESOURCE_UPDATED = "update"
    RESOURCE_DELETED = "delete"
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    RECONNECTED = "reconnected"
    INVALID_AUTH = "invalid_auth"



import json
import os

import pytest

from aioafero import device

current_path = os.path.dirname(os.path.realpath(__file__))


with open(os.path.join(current_path, "v1", "data", "device_lock.json")) as fh:
    device_lock_response = json.load(fh)
    lock_dev = device.get_afero_device(device_lock_response[0])


@pytest.mark.parametrize(
    "afero_device,expected",
    [
        # Everything is set correctly
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "model",
                "device_class": "device_class",
                "default_name": "default_name",
                "default_image": "default_image",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "model",
                    "device_class": "device_class",
                    "default_name": "default_name",
                    "default_image": "default_image",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [],
                }
            ),
        ),
        # DriskolFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-snyder-park-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "Driskol",
                    "device_class": "fan",
                    "default_name": "default_name",
                    "default_image": "ceiling-fan-snyder-park-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [],
                }
            ),
        ),
        # VinwoodFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-vinings-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "Vinwood",
                    "device_class": "fan",
                    "default_name": "default_name",
                    "default_image": "ceiling-fan-vinings-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [],
                }
            ),
        ),
        # ZandraFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "TBD",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-chandra-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "Zandra",
                    "device_class": "fan",
                    "default_name": "default_name",
                    "default_image": "ceiling-fan-chandra-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [],
                }
            ),
        ),
        # NevaliFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "TBD",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-ac-cct-dardanus-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "Nevali",
                    "device_class": "fan",
                    "default_name": "default_name",
                    "default_image": "ceiling-fan-ac-cct-dardanus-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [],
                }
            ),
        ),
        # TagerFan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "",
                "device_class": "fan",
                "default_name": "default_name",
                "default_image": "ceiling-fan-slender-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "Tager",
                    "device_class": "fan",
                    "default_name": "default_name",
                    "default_image": "ceiling-fan-slender-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [],
                }
            ),
        ),
        # Dimmer switch
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "dimmer",
                "device_class": "switch",
                "default_name": "default_name",
                "default_image": "ceiling-fan-slender-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        **{
                            "functionClass": "brightness",
                            "functionInstance": None,
                            "lastUpdateTime": 1668551478232,
                            "value": 40,
                        }
                    )
                ],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "dimmer",
                    "device_class": "light",
                    "default_name": "default_name",
                    "default_image": "ceiling-fan-slender-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [
                        device.AferoState(
                            **{
                                "functionClass": "brightness",
                                "functionInstance": None,
                                "lastUpdateTime": 1668551478232,
                                "value": 40,
                            }
                        )
                    ],
                }
            ),
        ),
        # Glass door
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "glass-door",
                "device_class": "glass-door",
                "default_name": "default_name",
                "default_image": "glass-door-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        **{
                            "functionClass": "power",
                            "functionInstance": None,
                            "lastUpdateTime": 1668551478232,
                            "value": "off",
                        }
                    )
                ],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "glass-door",
                    "device_class": "switch",
                    "default_name": "default_name",
                    "default_image": "glass-door-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [
                        device.AferoState(
                            **{
                                "functionClass": "power",
                                "functionInstance": None,
                                "lastUpdateTime": 1668551478232,
                                "value": "off",
                            }
                        )
                    ],
                    "manufacturerName": "Feather River Doors",
                }
            ),
        ),
        # Exhaust fan
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "glass-door",
                "device_class": "exhaust-fan",
                "default_name": "default_name",
                "default_image": "fan-exhaust-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        **{
                            "functionClass": "power",
                            "functionInstance": None,
                            "lastUpdateTime": 1668551478232,
                            "value": "off",
                        }
                    )
                ],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "BF1112",
                    "device_class": "exhaust-fan",
                    "default_name": "default_name",
                    "default_image": "fan-exhaust-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [
                        device.AferoState(
                            **{
                                "functionClass": "power",
                                "functionInstance": None,
                                "lastUpdateTime": 1668551478232,
                                "value": "off",
                            }
                        )
                    ],
                    "manufacturerName": None,
                }
            ),
        ),
        # 12A19060WRGBWH2
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "glass-door",
                "device_class": "light",
                "default_name": "default_name",
                "default_image": "a19-e26-color-cct-60w-smd-frosted-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        **{
                            "functionClass": "power",
                            "functionInstance": None,
                            "lastUpdateTime": 1668551478232,
                            "value": "off",
                        }
                    )
                ],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "12A19060WRGBWH2",
                    "device_class": "light",
                    "default_name": "default_name",
                    "default_image": "a19-e26-color-cct-60w-smd-frosted-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [
                        device.AferoState(
                            **{
                                "functionClass": "power",
                                "functionInstance": None,
                                "lastUpdateTime": 1668551478232,
                                "value": "off",
                            }
                        )
                    ],
                    "manufacturerName": None,
                }
            ),
        ),
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "glass-door",
                "device_class": "light",
                "default_name": "default_name",
                "default_image": "slide-dimmer-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        **{
                            "functionClass": "power",
                            "functionInstance": None,
                            "lastUpdateTime": 1668551478232,
                            "value": "off",
                        }
                    )
                ],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "HPDA110NWBP",
                    "device_class": "light",
                    "default_name": "default_name",
                    "default_image": "slide-dimmer-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [
                        device.AferoState(
                            **{
                                "functionClass": "power",
                                "functionInstance": None,
                                "lastUpdateTime": 1668551478232,
                                "value": "off",
                            }
                        )
                    ],
                    "manufacturerName": None,
                }
            ),
        ),
        (
            {
                "id": "id",
                "device_id": "device_id",
                "model": "TBD",
                "device_class": "switch",
                "default_name": "default_name",
                "default_image": "smart-switch-icon",
                "friendly_name": "friendly_name",
                "functions": ["functions!"],
                "states": [
                    device.AferoState(
                        **{
                            "functionClass": "power",
                            "functionInstance": None,
                            "lastUpdateTime": 1668551478232,
                            "value": "off",
                        }
                    )
                ],
            },
            device.AferoDevice(
                **{
                    "id": "id",
                    "device_id": "device_id",
                    "model": "HPSA11CWB",
                    "device_class": "switch",
                    "default_name": "default_name",
                    "default_image": "smart-switch-icon",
                    "friendly_name": "friendly_name",
                    "functions": ["functions!"],
                    "states": [
                        device.AferoState(
                            **{
                                "functionClass": "power",
                                "functionInstance": None,
                                "lastUpdateTime": 1668551478232,
                                "value": "off",
                            }
                        )
                    ],
                    "manufacturerName": None,
                }
            ),
        ),
    ],
)
def test_AferoDevice(afero_device, expected):
    assert device.AferoDevice(**afero_device) == expected


@pytest.mark.parametrize(
    "afero_device,expected_attrs",
    [
        # Validate when values are missing
        (
            {},
            {
                "id": None,
                "device_id": None,
                "model": None,
                "device_class": None,
                "default_name": None,
                "default_image": None,
                "friendly_name": None,
                "functions": [],
            },
        ),
        # Ensure values are properly parsed
        (
            device_lock_response[0],
            {
                "id": "5a5d5e04-a6ad-47c0-b9f4-b9fe5c049ef4",
                "device_id": "0123f95ec14bdb23",
                "model": "Keypad Deadbolt Lock",
                "device_class": "door-lock",
                "default_name": "Keypad Deadbolt Lock",
                "default_image": "keypad-deadbolt-lock-icon",
                "friendly_name": "Friendly Name 2",
                "functions": device_lock_response[0]["description"]["functions"],
            },
        ),
    ],
)
def test_get_afero_device(afero_device, expected_attrs):
    dev = device.get_afero_device(afero_device)
    for key, val in expected_attrs.items():
        assert (
            getattr(dev, key) == val
        ), f"Key {key} did not match, {getattr(dev, key)} != {val}"


@pytest.mark.parametrize(
    "data,expected_attrs",
    [
        # verify defaults
        (
            {
                "functionClass": "class",
                "value": "beans",
            },
            {
                "functionClass": "class",
                "value": "beans",
                "functionInstance": None,
                "lastUpdateTime": None,
            },
        ),
        # verify defaults
        (
            {
                "functionClass": "class",
                "value": "beans",
                "functionInstance": "beans",
                "lastUpdateTime": 4,
            },
            {
                "functionClass": "class",
                "value": "beans",
                "functionInstance": "beans",
                "lastUpdateTime": 4,
            },
        ),
    ],
)
def test_AferoState(data, expected_attrs):
    elem = device.AferoState(**data)
    for key, val in expected_attrs.items():
        assert getattr(elem, key) == val


def test_AferoDevice_hash():
    dev = device.get_afero_device(device_lock_response[0])
    hash_check = {dev: True}
    assert dev in hash_check


@pytest.mark.parametrize(
    "functions, func_class, func_instance, expected",
    [
        ([], "cool", "beans", None),
        (lock_dev.functions, "lock-pin", None, None),
        (lock_dev.functions, "lock-pin", "lock-pin-9", lock_dev.functions[19]),
    ],
)
def test_get_function_from_device(functions, func_class, func_instance, expected):
    assert (
        device.get_function_from_device(functions, func_class, func_instance)
        == expected
    )
